# Copyright (C) 2004  John Belmonte

"""Misc. support for clientlib

Dynamic objects:
    default_http_headers - default headers for all HTTP requests
"""

import time
import pickle
from httplib import HTTPConnection
import httplib
from urlparse import urlparse, urlunparse
from giftfile import http_status

_DUMP_HTTP = False


class HttpConnectionCache:
    """Cache of open HTTP connections"""

    def __init__(self):
        self.connections = {}

    def get_connection(self, host):
        if not self.connections.has_key(host):
            self.connections[host] = HTTPConnection(host)
            #self.connections[host].set_debuglevel(1)
        return self.connections[host]


_connection_cache = HttpConnectionCache()


class MemoryHttpCache(object):
    """Simple in-memory HTTP entity cache

    This is an example of an HTTP entity cache.  The only methods required
    for such an object are:

        set_content(url, entity, entity_tag, content_type)

        get_content(url) -> (entity, entity_tag, content_type)
            raises KeyError if item not in cache

    where "entity" is the HTTP response entity, "entity_tag" is the
    value of the Etag response header, and "content_type" is the
    Content-Type response header (may be None).

    Features of this implementation:
        * basic purging by time and total cache size
        * persist to a file
    """

    __slots__ = ['_map']

    def __init__(self):
        # map of url to (entity, entity_tag, content_type, access_time)
        self._map = {}

    def set_content(self, url, entity, entity_tag, content_type):
        self._map[url] = (entity, entity_tag, content_type, time.time())

    def get_content(self, url):
        """-> (entity, entity_tag, content_type)

        Updates access time.  Raises KeyError.
        """
        entity, entity_tag, content_type = self._map[url][:3]
        self.set_content(url, entity, entity_tag, content_type)
        return (entity, entity_tag, content_type)

    def _filter_newest_to_oldest(self, filter_function):
        items = self._map.items()
        # sort by access time, newest to oldest
        items.sort(lambda a, b: cmp(b[1][3], a[1][3]))
        self._map = dict(filter(filter_function, items))

    def purge_older_than(self, target_time):
        def f((url, (entity, tag, content_type, t))):
            return t > target_time
        self._filter_newest_to_oldest(f)

    def purge_to_total_size(self, target_bytes):
        class closure: bytes = 0
        def f((url, (entity, tag, content_type, t))):
            closure.bytes += len(entity)
            return closure.bytes < target_bytes
        self._filter_newest_to_oldest(f)

    def save_state(self, file):
        pickle.dump(self._map, file)

    def load_state(self, file):
        self._map = pickle.load(file)


default_http_headers = {}

class HttpResponse:
    """Proxy httplib.HTTPResponse

    The point of this proxy is to provide a single read of the response
    entity from the HTTP connection, and also to allow mutation of the
    response.  Mutation may be desirable for when providing transparent
    entity caching, for example.

    The basic httplib.HTTPResponse interface is implemented, with the
    following extensions:

        Attributes:
            raw_response - original httplib.HTTPResponse instance.  The
                response entity will already have been consumed.
            entity - HTTP response entity string
            is_cached_response - True if this response was provided by
                transparent caching

        Functions:
            getheaders() - sequence of header (name, value) tuples
    """
    def __init__(self, raw_response=None, extra_headers={}):
        if raw_response:
            to_copy = ['version', 'status', 'reason']
            map(lambda x: setattr(self, x, getattr(raw_response, x)), to_copy)
            self.raw_response = raw_response
            self.entity = raw_response.read()
        self._extra_headers = extra_headers
        self.is_cached_response = False

    def getheader(self, name, default=None):
        try:
            return self._extra_headers[name]
        except KeyError:
            return self.raw_response.getheader(name, default)

    def getheaders(self):
        return self.raw_response.msg.items() + self._extra_headers.items()

    def __str__(self):
        from giftfile._john_util import indent_text
        version = 'HTTP/%s.%s' % tuple(str(self.version))
        cached = ('', ' (cached response)')[self.is_cached_response]
        lines = [
            '%s %s %s%s' % (version, self.status, self.reason, cached),
        ]
        for (name, value) in self.getheaders():
            lines.append('  %s: %s' % (name.title(), value))
        lines.append('')
        if self.entity:
            if self.getheader('Content-Type') == 'application/xml':
                from giftfile._john_xml import parse_memory
                doc = parse_memory(self.entity)
                entity = doc.serialize(format=1)
            else:
                entity = self.entity
            lines.append(indent_text(entity, 2))
        return '\n'.join(lines)


def _dump_request(netloc, method, request_url, entity, headers):
    from giftfile._john_util import indent_text
    lines = [
        '%s %s %s' % (method, request_url, 'HTTP/1.1'),
        '  Host: %s' % netloc,
    ]
    for name, value in headers.iteritems():
        lines.append('  %s: %s' % (name.title(), value))
    lines.append('')
    if entity:
        if headers.get('Content-Type', None) == 'application/xml':
            from giftfile._john_xml import parse_memory
            doc = parse_memory(entity)
            entity = doc.serialize(format=1)
        lines.append(indent_text(entity, 2))
    return '\n'.join(lines)

def http_open(url, method='GET', body=None, headers={}, entity_cache=None):
    """-> HttpResponse instance

    Access HTTP resource via connection cache, with optional transparent
    entity caching.

    entity_cache: any object that implements HTTP cache interface described
        in _http_cache module, or None.  If the requested URL is in the
        cache, then the cached entity tag will be used to qualify the HTTP
        request with an If-None-Match.  If this results in a NOT_MODIFIED
        response, the returned response object will be adjusted to appear
        like a normal OK response.  Otherwise, assuming Etag is present,
        the entity cache will be updated with the new response.
    """
    is_cachable = (method == 'GET')
    all_headers = dict(default_http_headers)
    all_headers.update(headers)
    if is_cachable and entity_cache:
        try:
            cached_entity, cached_entity_tag, cached_type = \
                entity_cache.get_content(url)
            all_headers['If-None-Match'] = cached_entity_tag
        except KeyError:
            cached_entity = None
    (scheme, netloc, path, params, query, fragment) = urlparse(url)
    assert scheme == 'http'
    connection = _connection_cache.get_connection(netloc)
    # NOTE: HTTPConnection.request URL must not include scheme and server
    request_url = urlunparse((None, None, path, params, query, fragment))
    if _DUMP_HTTP:
        print _dump_request(netloc, method, request_url, body, all_headers)
    try:
        connection.request(method, request_url, body, all_headers)
        raw_response = connection.getresponse()
    except httplib.BadStatusLine:
        # This exception will happen if the server closed the persistent
        # connection.
        #print 'Re-opening connection to %s' % netloc
        connection.close()
        connection.request(method, request_url, body, all_headers)
        raw_response = connection.getresponse()
    # NOTE: we might use raw_response.will_close here, removing the connection
    #   from the cache if we know the server intends to close it.
    if raw_response.status == http_status.NOT_MODIFIED:
        assert cached_entity is not None
        #print 'using cached entity for "%s"' % url
        extra_headers = {'Content-Length': str(len(cached_entity))}
        if cached_type:
            extra_headers['Content-Type'] = cached_type
        response = HttpResponse(raw_response, extra_headers)
        response.entity = cached_entity
        response.status = http_status.OK
        response.reason = http_status.reasons[http_status.OK]
        response.is_cached_response = True
    else:
        response = HttpResponse(raw_response)
        entity_tag = response.getheader('Etag', None)
        content_type = response.getheader('Content-Type', None)
        if is_cachable and entity_cache and entity_tag:
            entity_cache.set_content(url, response.entity, entity_tag,
                content_type)
    if _DUMP_HTTP: print response
    return response

def http_open_xml(url, **keywords):
    """Get HTTP XML resource"""
    return http_open(url, headers={'Accept': 'application/xml'}, **keywords)

def http_xml_request(url, body, **keywords):
    """Post XML to HTTP resource"""
    return http_open(url, 'POST', body,
        headers={'Content-Type': 'application/xml'}, **keywords)

def _test_memory_http_cache():
    import os

    def assert_cache(cache, present=[], absent=[]):
        for url in present:
            if url not in cache._map:
                raise AssertionError('missing expected cache item "%s"' % url)
        for url in absent:
            if url in cache._map:
                raise AssertionError('unexpected cache item "%s"' % url)

    a = MemoryHttpCache()
    a.set_content('foo', 'content', 'tag', 'content_type');
    print a.get_content('foo')
    state_file = os.tmpfile()
    a.save_state(state_file)
    a = MemoryHttpCache()
    state_file.seek(0)
    a.load_state(state_file)
    assert_cache(a, ['foo'])
    time.sleep(.1)
    a.set_content('bar', 'content2', 'tag2', 'content_type2');
    assert_cache(a, ['foo', 'bar'])
    a.purge_older_than(0)
    assert_cache(a, ['foo', 'bar'])
    a.purge_to_total_size(999)
    assert_cache(a, ['foo', 'bar'])
    a.purge_older_than(time.time()-.05)
    assert_cache(a, ['bar'], ['foo'])
    a.set_content('baz', 'content3', 'tag3', 'content_type3');
    a.purge_to_total_size(len('content')*1.5)
    assert_cache(a, ['baz'], ['bar'])
    a.purge_to_total_size(len('content')*.5)
    assert_cache(a, absent=['baz'])
    print 'test complete'

def _main():
    _test_memory_http_cache()

if __name__ == '__main__':
    _main()
