# Copyright (C) 2004  Anthony Carrico

"""One ring to rule them all, One ring to find them, One ring to bring
them all and in the darkness bind them."""

import weakref

class _Id(object):
    _serial_number = 0
    def __init__(self):
        self.id = _Id._serial_number
        _Id._serial_number += 1

class ContextError(KeyError):
    def __init__(self):
        KeyError.__init__(self)

class Context(object):
    """A dynamic context.

    Sometimes you want a client to be able to customize an interface,
    an object, a procedure, a module, a protocol, or an interlocking
    collection of these--you want to give carry some state:
      * file handles
      * foreign (non-garbage-collected) objects
      * memoized computations
      * objects caches
      * precompiled regular expressions
      * precompiled schemas
      * custom initial values
      * choices/switches

    At the same time, you don't want to burden the client with
    managing your little state in addition to all the other states in
    the system. The dynamic context is an answer.

    You ask the client for a context, and you add your state to the
    context. You have just passed the management responsibility off to
    the client (be it another library, or an application). Your
    clients decide and control the states by deciding which when to
    create, use, and destroy contexts.

    This class is not thread safe."""

    def __init__(self):
        self.state = {}
        self.stack = []

    def __del__(self):
        stack = self.stack
        count = len(stack)
        while count > 0:
            stack[count - 1]()
            count-=1

    def add(self, key, value, unwinder=None):
        """Add some state to the dynamic context.

        Clients should make should manage keys carefully, since
        completely unrelated libraries can share a dynamic context.

        The value is anything, it might even be another context.

        The unwinder is a thunk (a procedure of zero arguments) that
        will be called when the dynamic context is exited."""

        self.state[key]=value
        if unwinder != None:
            self.stack.append(unwinder)

    def __getitem__(self, key):
        """Return the state associated with the given key.

        Raise ContextError if state hasn't been added under that key."""

        try:
            return self.state[key]
        except KeyError:
            raise ContextError

    def new_id():
        """Unique ids might be useful when creating keys."""
        return _Id()
    new_id=staticmethod(new_id)

class SharedResource(object):
    """A lazy shared resource.

    The resource is (re)created on demand, and destroyed when it is no
    longer referenced."""

    def __init__(self, make, open=None, close=None,):
        """Two disciplines are available for creating/destroying the
        resource:

          make() -> resource

        or

          token = open()
          make(token) -> resource
          ...
          close(token)"""

        self.id = Context.new_id()
        self.make = make
        self.open = open
        self.close = close
        self.ref = None
        if (((open == None) and (close != None))
            or ((open != None) and (close == None))):
            raise TypeError

    def _get(self):
        ref = self.ref
        if ref == None:
            resource = None
        else:
            resource = ref()
        if resource == None:
            if self.open==None:
                resource = self.make()
                self.ref= weakref.ref(resource)
            else:
                token = self.open()
                resource = self.make(token)
                self.ref= weakref.ref(resource, lambda ref: self.close(token))
        return resource

    def __call__(self, context=None):
        """Access the resource.

        If you supply an optional Context, then the context will
        reference the shared resource."""

        if context == None:
            return self._get()
        else:
            try:
                return context[self.id]
            except ContextError:
                resource = self._get()
                context.add(self.id, resource)
                return resource

def _examples():
    ############################################################
    _factorial_id = Context.new_id()

    def factorial(context, n):
        """Memoize the function in a context."""
        key = (_factorial_id, n)
        try:
            return context[key]
        except ContextError:
            if (n > 1):
                value = n * factorial(context, n-1)
            else:
                value = 1
            context.add(key, value)
            print 'factorial of ' + str(n) + '=' + str(value)
            return value

    ############################################################
    class Resource(object):
        def __init__(self):
            print "Resource.__init__"
        def close(self):
            print "Resource.close"

    class Foo(object):
        '''Share one resource per context.'''

        _resource_id = Context.new_id()

        def __init__(self, context):
            self.context = context

        def do(self):
            resource = self._get_resource()

        def _get_resource(self):
            '''Store the resource in the context.'''

            try:
                return self.context[Foo._resource_id]
            except KeyError:
                resource = Resource()
                self.context.add(Foo._resource_id,
                                 resource,
                                 lambda: resource.close())
                return resource

    ############################################################

    def open(name): print 'open' ; return 1
    def close(fd): print 'close'

    class Database(object):
        def __init__(self, fd):
            print 'Database.__init__'

    class Boo(object):
        '''Share one resource among several contexts.

        For variety, this time the context is provided to do instead
        of the constructor.'''

        _resource = SharedResource(lambda fd: Database(fd),
                                   lambda: open('text.txt'),
                                   lambda fd: close(fd))

        def __init__(self):
            pass

        def do(self, context=None):
            resource = Boo._resource(context)
            # ...

    ############################################################
    # import factorial...
    # import Foo...
    # import Boo...
    # Start a context:
    ctxt1 = Context()
    assert factorial(ctxt1, 3) == 6
    assert factorial(ctxt1, 1) == 1
    assert factorial(ctxt1, 1) == 1
    assert factorial(ctxt1, 1) == 1
    assert factorial(ctxt1, 3) == 6
    b1 = Boo()
    b1.do(ctxt1)
    # Start an overlapping context:
    ctxt2 = Context()
    assert factorial(ctxt2, 3) == 6
    b2 = Boo()
    b2.do(ctxt2)
    # End the first context:
    b1=None
    f1 = Foo(ctxt2)
    f1.do()
    b2.do(ctxt2)
    assert factorial(ctxt2, 2) == 2
    assert factorial(ctxt2, 1) == 1
    assert factorial(ctxt2, 3) == 6
    # End the second context:
    ctxt2 = None
    ############################################################

if __name__ == '__main__':
    _examples()
