# Copyright (C) 2004  Anthony Carrico

"""Data format/type utilities and bases."""

import funkxml
import libxml2
import sha
import base64
import binascii
import urllib2
import re

class Error(Exception):
    def __init__(self, message=''):
        Exception.__init__(self, message)

class XmlValidationError(Error):
    def __init__(self):
        Error.__init__(self)

def to_base64(buffer):
    """Return the base64 encoding of the given string.

    >>> to_base64('abc')
    'YWJj'
    """
    # Trim the trailing newline:
    return base64.encodestring(buffer)[:-1]

def from_base64(buffer):
    """Decode a base64 string.

    >>> from_base64('YWJj')
    'abc'
    """
    return base64.decodestring(buffer)

def sha1(buffer):
    """Return the sha1 digest of the given string."""
    return sha.new(buffer).digest()

def sha1_base64(buffer):
    """Return the base64 encoded sha1 hash of the given string.

    >>> sha1_base64('abc') == 'qZk+NkcGgWq6PiVxeFDCbJzQ2J0='
    True
    """
    return to_base64(sha1(buffer))

def fetch_sha1(uri):
    """Return the sha1 hash of the resource at the given uri."""
    pathname = libxml2.catalogResolveURI(uri)
    if pathname == None:
        file = urllib2.urlopen(uri)
    else:
        file = open(pathname)
    digester = sha.new()
    block = file.read(1024)
    while len(block) != 0:
        digester.update(block)
        block = file.read(1024)
    file.close()
    return digester.digest()

def fetch_sha1_base64(uri):
    """Return the base64 encoded sha1 hash of the resource at the given uri."""
    return to_base64(fetch_sha1(uri))

def to_funkxml_root_element(xml):
    """Return the root element in funkxml format."""

    if isinstance(xml, (str, unicode)):
        xml = funkxml.parse_memory(xml)
    elif isinstance(xml, libxml2.xmlNode):
        xml = funkxml.from_libxml2(xml)
    if isinstance(xml, funkxml.Root):
        return xml.element
    elif isinstance(xml, funkxml.Element):
        return xml
    else:
        raise TypeError

def hex_to_base64(data):
    return to_base64(binascii.a2b_hex(data))

def base64_to_hex(data):
    return binascii.b2a_hex(from_base64(data))

def hex_to_binary(data):
    return binascii.a2b_hex(data)

class XmlObject(object):
    """A base class for objects represented as xml elements."""

    funkxml = property(lambda self: self.element)
    libxml2 = property(lambda self: self.element.to_libxml2())

    def validate(self, context=None):
        """Validate the underlying xml

        An optional dynamic context can be supplied to optimize
        validation across instances validated in that context."""

        if not self.element.orphan_with_auto_ns_decls.relaxng_validate(
            self._schema_ctxt(context)):
            raise XmlValidationError
        return True

    def __init__(self, xml):
        """The xml source can be an unparsed xml string, a libxml2
        node, or a funkxml Node."""

        self.element = to_funkxml_root_element(xml)

def check_type(x, *types):
    if isinstance(x, types):
        return x
    else:
        raise TypeError

def check_string(s):
    if not isinstance(s, (str, unicode)):
        raise TypeError
    return s

_date_pattern = re.compile(r"""^(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})$""")

def check_date(date):
    match = _date_pattern.match(date)
    if ((match == None)
        or not (1 <= (int(match.group('month'))) <= 12)
        or not (1 <= (int(match.group('day'))) <= 31)):
        raise TypeError
    # ISSUE: could check if day matches year+month.
    return date

def check_uri(uri):
    return check_string(uri)

def optional(x):
    if len(x) == 1:
        return x[0]
    else:
        return None

def trap_key_error(thunk):
    try:
        return thunk()
    except KeyError:
        return None

def _test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    _test()
