# Copyright (C) 2004  John Belmonte

"""John's XML stuff"""

from types import StringTypes
import libxml2
import xml2tramp
from funkxml import Comment, Content
import funkxml
import xmldsig

def callable(obj):
    """A replacement for __builtin__.callable()

    When callable(obj) yields True it does not guarantee that calling the
    object will succeed.  An example is instances of Zope C extension
    classes such as ZODB Persistence.  See
    http://mail.zope.org/pipermail/zodb-dev/2003-October/006068.html.

    This replacement simply checks for the existence of a __call__ attribute,
    which seems to generally be more useful.  Note that it will not yield
    True for class objects.
    """
    return hasattr(obj, '__call__')

def get_xml_element(name, value=None, attributes=''):
    """Return XML element string.

    Synopsis:
        >>> get_xml_element('foo')
        '<foo/>'
        >>> get_xml_element('foo', 'bar')
        '<foo>bar</foo>'
        >>> get_xml_element('foo', attributes='bar="hi"')
        '<foo bar="hi"/>'

    ISSUE: this function requires value and attributes to be properly
        escaped, which is error prone
    """
    if len(attributes) > 0:
        attributes = ' ' + attributes
    if value:
        return '<%(name)s%(attributes)s>%(value)s</%(name)s>' % vars()
    else:
        return '<%(name)s%(attributes)s/>' % vars()

def list_to_xml(l):
    """Generate XML elements from a list

    This function generates XML elements from a list of
    (element_name, element_content) tuples.  The element_content value is
    normally a string, but some other types are supported:

        * If element_content is a list, list_to_xml will be called recursively
        to generate the element content.

        * If element_content is a function or other callable, it will be
        called with element_name as an argument.  Processing will continue as
        if element_content were the return value of the function.

        * For all other types, an attempt is made to convert the object to a
        string.

    A value of '' or None results in an empty element.

    Synopsis:

    >>> element_data = [
    ...     ('t1', 'hello'),
    ...     ('t2', None),
    ...     ('t3', lambda name : 'I am %s' % name),
    ...     ('ul', [('li', '1'), ('li', '2')])]
    >>> giftpool.util.list_to_xml(element_data)
    '<t1>hello</t1><t2/><t3>I am t3</t3><ul><li>1</li><li>2</li></ul>'

    TODO: support attributes
    """

    def generate_xml_element(element_item):
        name, contents = element_item
        if callable(contents):
            contents = contents(name)
            assert not callable(contents)
        if type(contents) is list:
            value = list_to_xml(contents)
        else:
            value = contents
        if value == None:
            value = ''
        return get_xml_element(name, str(value))

    element_values = map(generate_xml_element, l)
    return ''.join(element_values)


class Libxml2ErrorHandler:
    """
    This class is intended to patch up the sorry error handling system in
    libxml2.  The module's default error handling sends messages to standard
    out, which is very bad when we are in a CGI script.  Use this class
    to make the message part of libxml2's exceptions instead.

    See bug http://bugzilla.gnome.org/show_bug.cgi?id=122552.

    Usage synopsis:

        >>> libxml2_error_handler = Libxml2ErrorHandler()
        >>> xml = '<foo></bar>'
        >>> try:
        ...     libxml2.parseMemory(xml, len(xml))
        ... except libxml2.parserError, e:
        ...     raise libxml2_error_handler.convert_exception(e)
        ...
        Traceback (most recent call last):
          File "<stdin>", line 4, in ?
        libxml2.parserError: xmlParseMemory() failed
        Entity: line 1:
        error:
        Opening and ending tag mismatch: foo line 1 and bar

        <foo></bar>

                   ^

    NOTE: unregistering the handler must be done by the user
    """

    def __init__(self, register=True):
        self.clear_last_error()
        if register:
            libxml2.registerErrorHandler(self, None)

    def clear_last_error(self):
        """Clear last error message

        Since error callback returns error line-by-line, this function
        must be called to signal the end of an error message.
        """
        self.last_error = []

    def get_last_error(self):
        """Return last error message string"""
        return '\n'.join(self.last_error)

    def __call__(self, context, message):
        """Callback function for libxml2 error handling"""
        self.last_error.append(message)

    def convert_exception(self, exception):
        """Add full error message to a libxml2 exception"""
        exception.msg = exception.msg + '\n' + self.get_last_error()
        self.clear_last_error()
        return exception


_libxml2_error_handler = Libxml2ErrorHandler()

def convert_libxml2_exception(exception):
    """Add error details to a libxml2 exception"""
    return _libxml2_error_handler.convert_exception(exception)

def consume_libxml2_error():
    """Get last libxml2 error string and clear error state"""
    error = _libxml2_error_handler.get_last_error()
    _libxml2_error_handler.clear_last_error()
    return error


class XmlParseError(Exception):
    def __init__(self, message=''):
        Exception.__init__(self, message)

class XmlElementError(XmlParseError):
    def __init__(self, element_name):
        XmlParseError.__init__(self, 'node has no element "%s"' % element_name)

class XmlTraversalError(XmlParseError):
    def __init__(self, msg=''):
        XmlParseError.__init__(self, msg)

class XmlValidationError(Exception):
    def __init__(self, msg=''):
        Exception.__init__(self, msg)


class AutoXmlDoc(libxml2.xmlDoc):
    """libxml2.xmlDoc that frees itself on destruction

    An instance of this class is intended to be a drop-in replacement for a
    libxml2.xmlDoc.
    """

    def __init__(self, doc):
        """
        doc: libxml2.xmlDoc instance (ownership is transferred)
        """
        libxml2.xmlDoc.__init__(self, doc._o)

    def __del__(self):
        # If there is an exception in the initializer of a derived classes
        # before our __init__ is called, freeDoc will fail.
        try:
            self.freeDoc()
        except AttributeError:
            pass


def parse_memory(xml_text):
    """-> AutoXmlDoc instance"""
    try:
        doc = libxml2.parseMemory(xml_text, len(xml_text))
    except libxml2.parserError, e:
        raise XmlParseError(convert_libxml2_exception(e))
    return AutoXmlDoc(doc)

# alias
XmlDoc = parse_memory


class RngSchema:
    """"RelaxNG schema"""

    def __init__(self, url=None, xml_text=None):
        if url:
            assert not xml_text
            parser_context = libxml2.relaxNGNewParserCtxt(url)
        else:
            assert not url
            parser_context = libxml2.relaxNGNewMemParserCtxt(xml_text)
        self.schema = parser_context.relaxNGParse()

    def validate(self, xml_doc):
        schema_context = self.schema.relaxNGNewValidCtxt()
        if xml_doc.relaxNGValidateDoc(schema_context) is not 0:
            raise XmlValidationError(consume_libxml2_error())


class ElementMap(object):
    """Present xml2tramp element as recursive map of element names to values

    Synopsis:

    >>> x = ElementMap(xml2tramp.CachingElement('''
    ... <doc>
    ...     <foo>
    ...         <bar>hello</bar>
    ...         <bar2><baz>goodbye</baz></bar2>
    ...     </foo>
    ... </doc>
    ... '''))
    >>> print x
    {'foo': ...}
    >>> print x['foo']
    {'bar': 'hello', 'bar2': ...}
    >>> print x['foo']['bar']
    hello
    >>> for item in x['foo']: print item
    ...
    hello
    {'baz': 'goodbye'}
    >>>

    This interface does not support XML attributes or mixed content.
    """
    __slots__ = ['_element', '_root']

    def __init__(self, xml2tramp_element, _root=None):
        self._element = xml2tramp_element
        self._root = _root or xml2tramp_element

    def __repr__(self):
        """-> verbose representation"""
        return '%s(%s)' % (xml2tramp._qualified_class_name(self),
            repr(self._element))

    def __str__(self):
        """-> dictionary representation"""
        def get_name(element):
            namespace, name = element.name
            if namespace == element.default_namespace:
                return name
            else:
                return element.name
        def item_string(element):
            if element.elements:
                return '%s: ...' % repr(get_name(element))
            else:
                return '%s: %s' % (repr(get_name(element)),
                    repr(element.content))
        s = map(item_string, self._element.elements)
        return '{%s}' % ', '.join(s)

    def __getitem__(self, index):
        """-> child ElementMap or contents"""
        child_element = self._element[index]
        if child_element.elements:
            # in case the child outlives us, have it hold a reference to root
            return self.__class__(child_element, _root=self._root)
        else:
            return child_element.content

    def __len__(self):
        """-> number of child elements"""
        return len(self._element.elements)

    def __contains__(self, index):
        """-> True if named child element exists"""
        return self._element.has_element(index)


def sign_xml(xml, pgp_ctxt, fingerprint=None):
    """-> XML text with enveloped PGP signature

    xml: input XML text
    pgp_ctxt: openpgp.Context instance
    fingerprint: PGP fingerprint of signer (if None, keyring default is used)

    The signature is added after the last child of the root element.
    """

    if fingerprint:
        signers = [fingerprint]
    else:
        signers = None

    sig_ctxt = xmldsig.enveloped_openpgp_custom(pgp_ctxt)

    old_content = funkxml.parse_memory(xml).element.content
    new_content = old_content + Content((Comment('sig goes here.'),))
    bogus_sig = old_content.parent.with_content(new_content).content[-1]

    sig = sig_ctxt.sign(bogus_sig, signers=signers)

    return sig.top.serialize()


class SignatureInfo(object):
    """XML digital signature information

    attributes:
        sigs - list of signatures (see openpgp.Context.detach_verify)
        message_digest - binary message digest
    """
    __slots__ = ['sigs', 'message_digest']

def validate_signed_xml(xml, pgp_ctxt, validate_xml=True):
    """-> SignatureInfo instance

    xml: XML text or libxml2 doc-like instance with enveloped PGP signature
    pgp_ctxt: openpgp.Context instance

    The signature is assumed to be the last child of the root element.

    TODO: accept signature element at any location
    """

    sig_ctxt = xmldsig.enveloped_openpgp_custom(pgp_ctxt)

    if isinstance(xml, StringTypes):
        sig = funkxml.parse_memory(xml).element.content[-1]
    elif isinstance(xml, funkxml.Root):
        sig = xml.element.content[-1]
    elif isinstance(xml, funkxml.Element):
        sig = xml.content[-1]
    elif xml.type == 'element':
        sig = xml.last
    elif xml.type == 'document_xml':
        sig = xml.children.last
    else:
        assert False

    info = SignatureInfo()
    info.sigs = sig_ctxt.validate(sig, validate_xml=validate_xml)
    info.message_digest = xmldsig.get_message_digest(sig)

    return info
