# Copyright (C) 2004  Anthony Carrico

"""OS utilities

Module attributes:
    nullout - Python file object for writing to "/dev/null" or equivalent
    null_file_name - path of null file
"""

import os
import select

class ChildPipe(object):
    """Convenience wrapper for os.pipe"""

    def __init__(self, child_fd=None, parent_fd=None):
        self._child_fd = child_fd
        self._parent_fd = parent_fd
        self._read_fd, self._write_fd = os.pipe()

    def _maybe_change_fd(src, dest):
        if dest:
            os.dup2(src, dest)
            os.close(src)
            return dest
        else:
            return src
    _maybe_change_fd = staticmethod(_maybe_change_fd)

class PipeToChild(ChildPipe):
    """Convenience wrapper for os.pipe from parent to child"""
    def __init__(self, child_fd=None, parent_fd=None):
        ChildPipe.__init__(self, child_fd, parent_fd)
    def extract_child_fd(self):
        os.close(self._write_fd)
        return self._maybe_change_fd (self._read_fd, self._child_fd)
    def extract_child_file(self):
        return os.fdopen(self.extract_child_fd(), 'r')
    def extract_parent_fd(self):
        os.close(self._read_fd)
        return self._maybe_change_fd (self._write_fd, self._parent_fd)
    def extract_parent_file(self):
        return os.fdopen(self.extract_parent_fd(), 'w')

class PipeFromChild(ChildPipe):
    """Convenience wrapper for os.pipe from child to parent"""
    def __init__(self, child_fd=None, parent_fd=None):
        ChildPipe.__init__(self, child_fd, parent_fd)
    def extract_child_fd(self):
        os.close(self._read_fd)
        return self._maybe_change_fd (self._write_fd, self._child_fd)
    def extract_child_file(self):
        return os.fdopen(self.extract_child_fd(), 'w')
    def extract_parent_fd(self):
        os.close(self._write_fd)
        return self._maybe_change_fd (self._read_fd, self._parent_fd)
    def extract_parent_file(self):
        return os.fdopen(self.extract_parent_fd(), 'r')

def offset_write (fd, str, start, end):
    """Like os.write, but with a start and end.

    Return the number of bytes actually written."""
    # ISSUE: Does slice copy? Is there a way to avoid the unnecessary copy?
    return os.write(fd, str[start:end])

def read_and_write_fds (read_fds, write_table):
    """Push and pull data from several file descriptors.

    write_table maps from an fd to outgoing string.

    Keep trying to read and write the given fds until all the data in
    the write_table are sent (closing the corresponding fds), and all
    the data in the read_fds are received (closing the corresponding
    fds). Return a read table mapping fd to incoming string."""

    block_size = 1024
    # Don't mutate the original read_fds:
    read_fds = read_fds[:]
    read_table = {}
    for fd in read_fds:
        read_table[fd]=[]
    offset_table = {}
    write_fds = []
    for fd in write_table:
        data = write_table[fd]
        if len(data)>0:
            offset_table[fd]=0
            write_fds.append(fd)
    while read_fds or write_fds:
        ready_read_fds, ready_write_fds, ignore = select.select (read_fds,
                                                                 write_fds,
                                                                 [])
        # Service the ready read fds:
        for fd in ready_read_fds:
            block = os.read(fd, block_size)
            if (len(block) == 0):
                os.close(fd)
                read_fds.remove(fd)
            else:
                read_table[fd].append(block)
        # Service the ready write fds:
        for fd in ready_write_fds:
            offset = offset_table[fd]
            data = write_table[fd]
            count = offset_write(fd, data, offset, min (offset + block_size, len(data)))
            offset = offset + count
            if (offset < len(data)):
                offset_table[fd] = offset
            else:
                os.close(fd)
                write_fds.remove(fd)
    for fd in read_table:
        read_table[fd] = "".join(read_table[fd])
    return read_table

def is_child(pid): return (pid == 0)


# Python library doesn't offer a portable /dev/null.
# See <http://mail.python.org/pipermail/python-list/2004-March/213590.html>.

if os.name in ['nt']:
    null_file_name = 'nul:'
else:
    null_file_name = '/dev/null'

nullout = file(null_file_name, 'w')
