# Copyright (C) 2004  Anthony Carrico

from giftfile._data_util import hex_to_base64, \
     hex_to_binary, to_base64, check_uri, check_string, check_type, \
     to_funkxml_root_element, XmlObject, trap_key_error, from_base64
from giftfile import funkxml
from giftfile.funkxml import Attribute, Element, NsDecl
from giftfile import xmldsig
from giftfile.xmldsig import KeyInfo, Reference, Signature, \
     get_ns_decl_ds, enveloped_openpgp_custom, \
     SignatureDigestMismatch
from giftfile.time_util import get_date_and_time, get_seconds
from giftfile._context import Context, SharedResource
from giftfile import relaxng
from giftfile import openpgp

from sets import Set

class Error(Exception):
    def __init__(self, message=''):
        Exception.__init__(self, message)

_ds = get_ns_decl_ds()
_cert = NsDecl(None, 'http://giftfile.org/ns/giftfile/0.0')

class Identity(object):
    def __init__(self, key_info, uri=None):
        self.key_info=check_type(key_info, KeyInfo)
        if uri != None: check_uri(uri)
        self.uri=uri

    def to_funkxml(self, prefix, name):
        if (self.uri != None):
            uri = Attribute(None, 'URI', self.uri)
        else:
            uri = None
        return Element.auto(prefix, name, self.key_info.funkxml, uri)

    def make_from_funkxml(element):
        return Identity(
            KeyInfo.make_from_funkxml(element.content_map[_ds('KeyInfo')][0]),
            trap_key_error(lambda: element.attrs['URI']).string_value)
    make_from_funkxml = staticmethod(make_from_funkxml)

    def sign(self, sig_ctxt, sig_placeholder):
        return sig_ctxt.sign(sig_placeholder, signers = [self.key_info])

    def verify(self, sig_ctxt, sig):
        try:
            openpgp_sigs = sig_ctxt.validate(sig, False)
            for openpgp_sig in openpgp_sigs:
                if isinstance(openpgp_sig, openpgp.SigGood):
                    if (hex_to_binary(openpgp_sig.fingerprint)
                        == self.key_info.binary):
                        if (openpgp_sig.expired
                            or openpgp_sig.expired_key
                            or openpgp_sig.revoked_key
                            or (openpgp_sig.trust == 'TRUST_NEVER')):
                            return False
                        else:
                            return True
                elif isinstance(opengpg_sig, openpgp.SigError):
                    pass
                elif isinstance(opengpg_sig, openpgp.SigBad):
                    pass
                else:
                    assert False
            return False
        except SignatureDigestMismatch:
            return false

class Identifier(object):
    def __init__(self, value, uri=None):
        self.value=check_string(value)
        if uri != None: check_uri(uri)
        self.uri=uri

    def to_funkxml(self, prefix, name):
        if (self.uri != None):
            uri = Attribute(None, 'URI', self.uri)
        else:
            uri = None
        return Element.auto(prefix, name, self.value, uri)

    def make_from_funkxml(element):
        return Identifier(element.string_value,
                          trap_key_error(lambda: element.attrs['URI']).string_value)
    make_from_funkxml = staticmethod(make_from_funkxml)

class PublicDomain(object):
    funkxml = property(
        lambda self: Element.auto(
        None, _cert('PublicDomain'),
        Element.auto(None, _cert('Year'), self.year),
        Element.auto(None, _cert('Author'), self.author)))

    def __init__(self, year, author):
        self.year=check_string(year)
        self.author=check_string(author)

    def get_license_uris(self):
        return Set()

class Copyright(object):
    funkxml = property(
        lambda self: Element.auto(
        None, _cert('Copyright'),
        Element.auto(None, _cert('Year'), self.year),
        Element.auto(None, _cert('Claimant'), self.claimant)))

    def __init__(self, year, claimant):
        self.year=check_string(year)
        self.claimant=check_string(claimant)

    def make_from_funkxml(element):
        return Copyright(
            element.content_map[_cert('Year')][0].string_value,
            element.content_map[_cert('Claimant')][0].string_value)
    make_from_funkxml = staticmethod(make_from_funkxml)

class License(object):
    funkxml = property(
        lambda self: Element.auto(
        None, _cert('License'),
        Attribute(None, 'URI', self.uri),
        map(lambda r: r.funkxml, self.copyrights)))

    def __init__(self, uri, copyrights):
        self.uri=check_uri(uri)
        for copyright in copyrights:
            check_type(copyright, Copyright)
        assert (len(copyrights) > 0)
        self.copyrights=copyrights

    def make_from_funkxml(element):
        return License(
            element.attrs['URI'].string_value,
            map(Copyright.make_from_funkxml,
                element.content_map[_cert('Copyright')]))
    make_from_funkxml = staticmethod(make_from_funkxml)

    def get_license_uris(self):
        return Set([self.uri])

class RightsAnd(object):
    funkxml = property(
        lambda self: Element.auto(
        None, _cert('And'),
        map(lambda r: r.funkxml, self.elements)))

    def __init__(self, elements):
        self.elements = _check_rights_elements(elements)

    def make_from_funkxml(element):
        elements = []
        for child in element.content:
            if isinstance(child, Element):
                elements.append(
                    _make_rights_element_from_funkxml(child))
        return RightsAnd(elements)
    make_from_funkxml = staticmethod(make_from_funkxml)

    def get_license_uris(self):
        result = Set()
        for element in self.elements:
            result |= element.get_license_uris()
        return result

class RightsOr(object):
    funkxml = property(
        lambda self: Element.auto(
        None, _cert('Or'),
        map(lambda r: r.funkxml, self.elements)))

    def __init__(self, elements):
        self.elements = _check_rights_elements(elements)

    def make_from_funkxml(element):
        elements = []
        for child in element.content:
            if isinstance(child, Element):
                elements.append(
                    _make_rights_element_from_funkxml(child))
        return RightsOr(elements)
    make_from_funkxml = staticmethod(make_from_funkxml)

    def get_license_uris(self):
        result = Set()
        for element in self.elements:
            result |= element.get_license_uris()
        return result

def _check_rights_element(element):
    check_type(element, PublicDomain, License, RightsAnd, RightsOr)
    return element

def _check_rights_elements(elements):
    for element in elements: _check_rights_element(element)
    return elements

_rights_element_table = {
    _cert('PublicDomain'):PublicDomain,
    _cert('License'):License,
    _cert('And'):RightsAnd,
    _cert('Or'):RightsOr
    }

def _make_rights_element_from_funkxml(element):
    return _rights_element_table[element.name].make_from_funkxml(element)

class Rights(object):
    funkxml = property(
        lambda self: Element.auto(
        None, _cert('Rights'),
        self.element.funkxml))

    def __init__(self, element):
        self.element =  _check_rights_element(element)

    def make_from_funkxml(element):
        for child in element.content:
            if isinstance(child, Element):
                return Rights(_make_rights_element_from_funkxml(child))
        assert False
    make_from_funkxml = staticmethod(make_from_funkxml)

    def get_license_uris(self):
        return self.element.get_license_uris()

class Donation(object):
    funkxml = property(
        lambda self: Element.auto(
        None, _cert('SuggestedDonation'),
        Element.auto(None, _cert('Value'), self.value),
        Element.auto(None, _cert('Currency'), self.currency)))

    def __init__(self, value, currency):
        self.value = check_string(value)
        self.currency = check_string(currency)

    def make_from_funkxml(element):
        return Donation(
            element.content_map[_cert('Value')][0].string_value,
            element.content_map[_cert('Currency')][0].string_value)
    make_from_funkxml = staticmethod(make_from_funkxml)

def _check_reference(reference):
    check_type(reference, Reference)
    assert reference.transforms == None
    assert reference.digest_method == xmldsig._algo_id_sha1
    assert reference.uri != None
    check_uri(reference.uri)
    assert reference.id == None
    assert reference.type == None
    return reference

class Certificate(XmlObject):
    """A giftfile certificate."""

    publisher = property(
        lambda self: Identity.make_from_funkxml(
        self.element.content_map[_cert('Publisher')][0]),
        doc='An Identity instance')

    giftpool = property(
        lambda self: trap_key_error(
        lambda: Identity.make_from_funkxml(
        self.element.content_map[_cert('Giftpool')][0])),
        doc='An Identity instance or None')

    source = property(
        lambda self: Reference.make_from_funkxml(
        self.element.content_map[_cert('Work')][0].
        content_map[_cert('Source')][0].
        content_map[_ds('Reference')][0]),
        doc='a Reference instance')

    variation = property(
        lambda self: trap_key_error(
        lambda: Reference.make_from_funkxml(
        self.element.content_map[_cert('Work')][0].
        content_map[_cert('Variation')][0].
        content_map[_ds('Reference')][0])),
        doc='a Reference instance or None')

    binary = property(
        lambda self: trap_key_error(
        lambda: Reference.make_from_funkxml(
        self.element.content_map[_cert('Work')][0].
        content_map[_cert('Binary')][0].
        content_map[_ds('Reference')][0])),
        doc='a Reference instance or None')

    series = property(
        lambda self: trap_key_error(
        lambda: Identifier.make_from_funkxml(
        self.element.content_map[_cert('Series')][0])),
        doc='An Identifier instance or None')

    issued = property(
        lambda self: get_seconds(
        self.element.content_map[_cert('Issued')][0].string_value),
        doc='a UTC date and time in seconds (integer).')

    name = property(
        lambda self:
        self.element.content_map[_cert('Work')][0]
        .content_map[_cert('Name')][0].string_value,
        doc='name (string).')

    rights = property(
        lambda self:
        Rights.make_from_funkxml(
        self.element.content_map[_cert('Work')][0].
        content_map[_cert('Rights')][0]),
        doc='a Rights instance')

    suggested_donation = property(
        lambda self:
        Donation.make_from_funkxml(
        self.element.content_map[_cert('Work')][0].
        content_map[_cert('SuggestedDonation')][0]),
        doc='a Donation instance')

    signature = property(
        lambda self:
        Signature(self.element.content_map[_ds('Signature')][0]),
        doc='a Signature instance')

    def make(sig_ctxt,
             source,
             name,
             rights,
             publisher,
             issued=None,
             work_uri=None,
             variation=None,
             binary=None,
             suggested_donation=None,
             series=None,
             giftpool=None):
        """Make a Certificate.

        Pass source as a Reference instance.
        Pass name as a string.
        Pass rights as a Rights instance.
        Pass publisher as an Identity instance.
        Pass issued as a UTC date and time in seconds (integer) or None.
        Pass work_uri as a string or None.
        Pass variation as a Reference instance or None.
        Pass binary as a Reference instance or None.
        Pass suggested_donation as a Donation instance or None.
        Pass series as an Identifier instance or None.
        Pass giftpool as an Identity instance or None.
        """

        if work_uri != None:
            work_uri = Attribute(None, 'URI', check_uri(work_uri))
        if variation != None:
            variation = Element.auto(None, _cert('Variation'),
                                     _check_reference(variation).funkxml)
        if binary != None:
            binary = Element.auto(None, _cert('Binary'),
                                  _check_reference(binary).funkxml)
        if suggested_donation != None:
            suggested_donation = check_type(
                suggested_donation, Donation).funkxml
        if series != None:
            series = check_type(
                series, Identifier).to_funkxml(None, _cert('Series'))
        if giftpool != None:
            giftpool = check_type(
                giftpool, Identity).to_funkxml(None, _cert('Giftpool'))

        certificate = Element.auto(
            None, _cert('Certificate'),
            Element.auto(None, _cert('Issued'),
                         get_date_and_time(issued)),
            Element.auto(None, _cert('Work'),
                         work_uri,
                         Element.auto(None, _cert('Name'),
                                      check_string(name)),
                         Element.auto(None, _cert('Source'),
                                      _check_reference(source).funkxml),
                         variation,
                         binary,
                         check_type(rights, Rights).funkxml,
                         suggested_donation),
            series,
            check_type(publisher, Identity).to_funkxml(
            None, _cert('Publisher')),
            giftpool,
            Element.auto(None, _ds('Signature'))).orphan_with_auto_ns_decls
        sig = publisher.sign(
            sig_ctxt, certificate.content_map[_ds('Signature')][0])
        return Certificate(sig.parent)
    make = staticmethod(make)

    _schema_ctxt = SharedResource(lambda: relaxng.fetch_ctxt(
        'http://giftfile.org/2004/02/certificate.rng'))

    def __init__(self, xml):
        XmlObject.__init__(self, xml)

    def verify_signature(self, sig_ctxt):
        return self.publisher.verify(
            sig_ctxt,
            self.funkxml.content_map[_ds('Signature')][0])

def _example():
    from os.path import expanduser, exists
    from giftfile import openpgp

    # NOTE: Use a bogus keyring if available:
    testing_gpg_homedir = expanduser('~/tmp/gpg')
    if exists(testing_gpg_homedir):
        gpg_homedir = testing_gpg_homedir
    else:
        gpg_homedir = None

    pgp_ctxt = openpgp.Context(homedir=gpg_homedir)
    fingerprints = pgp_ctxt.list_keys(secret_keys=True)
    assert len(fingerprints) > 0
    sig_ctxt = enveloped_openpgp_custom(pgp_ctxt)

    giftpool = Identity(
        KeyInfo.make_from_hex(
        "44f91abd6dc120cf8a33e11459a6d055e23478ce"),
        "http://test-giftpool.giftfile.org/")

    publisher = Identity(
        KeyInfo.make_from_hex(fingerprints[0]),
        "http://foofactory.example/")

    series = Identifier(
        "SuperFoo", "http://foofactory.example/products#SuperFoo")

    work_uri = "http://foofactory.example/releases/SuperFoo#0.92"

    source_uri = "http://foofactory.example/files/superfoo_0.92.orig.tar.gz"
    source_file = "file:certificate.py"    # <- Not really
    source_ref = Reference.fetch_and_make(source_file, uri=source_uri)

    variation_uri = "http://foofactory.example/files/superfoo_0.92-5.diff.gz"
    variation_file = "file:certificate.py" # <- Not really
    variation_ref = Reference.fetch_and_make(variation_file, uri=variation_uri)

    binary_uri = "http://foofactory.example/files/superfoo_0.92-5_i386.deb"
    binary_file = "file:certificate.py"    # <- Not really
    binary_ref = Reference.fetch_and_make(binary_file, uri=binary_uri)

    name = "SuperFoo version 0.92"
    license_uri = "http://giftfile.org/known_licenses#MIT"
    rights = Rights(License(license_uri,
                            [Copyright("2004", "Fracis Foo III")]))

    suggested_donation = Donation("0.75", "USD")

    cert = Certificate.make(
        sig_ctxt,
        source_ref,
        name,
        rights,
        publisher,
        work_uri=work_uri,
        variation=variation_ref,
        binary=binary_ref,
        suggested_donation=suggested_donation,
        series=series,
        giftpool=giftpool)

    assert cert.validate()
    assert cert.verify_signature(sig_ctxt)
    assert cert.publisher.key_info.binary == publisher.key_info.binary
    assert cert.publisher.uri == publisher.uri
    assert cert.giftpool.key_info.binary == giftpool.key_info.binary
    assert cert.giftpool.uri == giftpool.uri
    assert cert.source.uri == source_uri
    assert (from_base64(cert.source.digest_value)
            == from_base64(source_ref.digest_value))
    assert cert.variation.uri == variation_uri
    assert (from_base64(cert.variation.digest_value)
            == from_base64(variation_ref.digest_value))
    assert cert.binary.uri == binary_uri
    assert (from_base64(cert.binary.digest_value)
            == from_base64(binary_ref.digest_value))
    assert cert.series.value == series.value
    assert cert.series.uri == series.uri
    assert isinstance(cert.issued, int)
    assert cert.name == name
    assert cert.rights.element.uri == rights.element.uri
    assert (cert.rights.element.copyrights[0].year
            == rights.element.copyrights[0].year)
    assert (cert.rights.element.copyrights[0].claimant
            == rights.element.copyrights[0].claimant)
    assert cert.suggested_donation.value == suggested_donation.value
    assert cert.suggested_donation.currency == suggested_donation.currency

    assert cert.rights.get_license_uris() == Set([license_uri])

    assert (
        Set(['http://a.example', 'http://b.example', 'http://c.example']) ==
        Rights(RightsAnd
               ((License('http://a.example',
                         [Copyright('1900', 'example')]),
                 License('http://b.example',
                         [Copyright('1900', 'example')]),
                 PublicDomain('1900', 'example'),
                 RightsOr((License('http://a.example',
                                   [Copyright('1900', 'example')]),
                           License('http://c.example',
                                   [Copyright('1900', 'example')]),
                           PublicDomain('1900', 'example'))))))
        .get_license_uris())

    print cert.funkxml.serialize(indent=True)

if __name__ == '__main__':
    funkxml.init_libxml2_debug()
    _example()
    funkxml.quit_libxml2_debug()
