# Copyright (C) 2004  John Belmonte

"""Giftpool client library

This is the lowest-level interface for communicating with a giftpool.

The fact that giftpool communication uses HTTP and XML are details of
the protocol design.  This module exposes those details.  For example,
you can inspect the HTTP headers or XML document of a server response.
The layout of each response record must be known by the user.  Fields
in the records are not converted to native types.

Within the domain of HTTP and XML, however, the interface provides a
number of services:

    * basic XML parsing of incoming messages
    * automatic checking of HTTP response status
    * digital signing of outgoing requests
    * verification of giftpool digital signatures on incoming messages
    * persistent HTTP connections
    * caching of HTTP connections
    * fully synchronized transparent caching of HTTP response entities

The main class in this module is GiftpoolInterface.  Everything else is
used as input or output of that object.

The following are considered outside the scope of this library:
    * sending user key to the giftpool's keyserver
    * tracking of transactions
    * resubmitting an individual allocation when the server didn't
        recognize the giftfile ID

Dynamic objects provided by this module:
    default_http_headers - default headers for all HTTP requests
    pgp_log_file - Python file object for PGP log (defaults to stderr).
        This should be set before using the module.


Synopsis:
    >>> from giftfile import clientlib, transaction_status
    >>> giftpool_url = 'http://example.net/giftpool/'
    >>> giftpool_fingerprint = 'bc0f2d9f5d09a55e6aa291988581fe54d9c16916'
    >>> giftpool_id = clientlib.Identity(giftpool_fingerprint)
    >>> http_cache = clientlib.MemoryHttpCache()
    >>> giftpool = clientlib.GiftpoolInterface(giftpool_url, giftpool_id,
    ...     http_cache)
    >>> info_record = giftpool.get_info().element_map
    >>> info_record['name']
    "John's Giftpool"
    >>> info_record['donation_friction']
    '0.25'
    >>> print info_record
    {'uri': 'http://example.net/giftpool/', 'name': "John's Giftpool", ...}
    >>> user_fingerprint = '026f68efbf3b354d38c81ebe54905d336a7a12e3'
    >>> user_id = clientlib.Identity(user_fingerprint)
    >>> donation_request = giftpool.create_donation(user_id,
    ...     clientlib.get_date_and_time())
    >>> donation_url = donation_request.send().url
    >>> donation_url
    'http://example.net/giftpool/1/private/donations/012bdc9c...'
    >>> donation_record = giftpool.get_donation(donation_url).element_map
    >>> assert transaction_status.get_code(donation_record['status']) == \
    ...     transaction_status.PAYMENT_NOT_RECEIVED
    >>> donation_record['payment_id']
    '5b751cd8'
    >>> user_record = giftpool.get_user(donation_record['user']).element_map
    >>> user_url = giftpool.locate_user_record(user_record['id']).url
    >>> user_url
    'http://example.net/giftpool/1/private/users/026f68ef...'
    >>> assert user_record['uri'] == user_url
    >>>
"""
# TODO: let the outside world know when a signature is valid, but untrusted
# TODO: in giftpool_record, verify that ID matches signature
# ISSUE: transparent caching is probably better left to an external program

import sys
from urllib import urlencode
from giftfile import http_status, transaction_status, openpgp, xml2tramp
from giftfile.defs import GIFTPOOL_NAMESPACE as XML_NAMESPACE
from giftfile.time_util import get_date, get_date_and_time
from giftfile._john_xml import list_to_xml, get_xml_element, sign_xml, \
    validate_signed_xml, ElementMap, parse_memory
from giftfile._john_util import indent_text
# Local
from _clientlib_util import http_open_xml, http_xml_request, \
    default_http_headers, MemoryHttpCache

__version__ = '0.0'
#__revision__ = '$Revision: 0 $'[11:-2]
__author__ = 'John Belmonte <john@neggie.net>'

_USER_AGENT = 'giftpool_clientlib/%s' % __version__

default_http_headers['User-Agent'] = _USER_AGENT
pgp_log_file = sys.stderr

class ClientlibError(Exception):
    def __init__(self, message=''):
        Exception.__init__(self, message)

class UnexpectedHttpResponse(ClientlibError):

    # these headers will be displayed in exception message
    _interesting_headers = (
        'retry-after',
    )

    def __init__(self, response, response_entity, expected_status):
        """
        response: httplib.HTTPResponse instance
        """
        self.response = response
        collected_headers = []
        for (name, value) in self.get_header_items():
            if name.lower() in self._interesting_headers:
                collected_headers.append('%s: %s' % (name.title(),
                    value))
        header_info = ''
        if collected_headers:
            header_info = 'Headers of interest:\n%s\n' % \
                indent_text('\n'.join(collected_headers), 4)
        message = '%d %s (expected %d)\n%s%s' % (response.status,
            response.reason, expected_status, header_info,
            response_entity)
        ClientlibError.__init__(self, message)

    def get_status(self):
        """-> HTTP status code integer"""
        return self.response.status

    def get_reason(self):
        """-> HTTP status reason string"""
        return self.response.reason

    def get_header(self, name, default=None):
        return self.response.getheader(name, default)

    def get_header_items(self):
        """-> list of (name, value) pairs"""
        # ISSUE: HTTPResponse does not publish a way to get all headers
        return self.response.msg.items()

class SignatureVerificationFailed(ClientlibError):
    def __init__(self, msg=''):
        ClientlibError.__init__(self, msg)


def _check_http_response(response, expected_status=http_status.OK):
    """Check HTTP response and return response entity string"""
    if response.status != expected_status:
        raise UnexpectedHttpResponse(response.raw_response, response.entity,
            expected_status)
    return response.entity

def _check_signature(doc, identity):
    """Verify XML signature against identity

    If identity is an AutoIdentity instance without its fingerprint attribute,
    the fingerprint will be filled in according to the signature.
    """

    sigs = validate_signed_xml(doc, identity.pgp_context).sigs
    if isinstance(identity, AutoIdentity) and not identity.fingerprint:
        assert len(sigs) == 1
        sig = sigs[0]
        if isinstance(sig, openpgp.SigGood):
            identity.fingerprint = sig.fingerprint.lower()
            return
    else:
        for sig in sigs:
            if isinstance(sig, openpgp.SigGood) and \
                    sig.fingerprint.lower() == identity.fingerprint.lower():
                return
    raise SignatureVerificationFailed('signatures: %s' % sigs)


class Identity(object):
    """User/Giftpool identity"""

    def __init__(self, fingerprint, pgp_homedir=None):
        self.fingerprint = fingerprint
        self.pgp_context = openpgp.Context(homedir=pgp_homedir,
            log_file=pgp_log_file)


class AutoIdentity(object):
    """Auto giftpool identity

    This is an identity that is incomplete at the time of instantiation,
    and will be filled in later.
    """

    def __init__(self, pgp_homedir=None):
        self.fingerprint = None
        self.pgp_context = openpgp.Context(homedir=pgp_homedir,
            log_file=pgp_log_file)


class GiftfileReference(object):
    """Reference to a giftfile (by ID)"""

    __slots__ = ['giftfile_id']

    def __init__(self, giftfile_id):
        self.giftfile_id = giftfile_id

    def _get_xml(self):
        return ('giftfile_id', self.giftfile_id)


class Certificate(object):
    """Giftfile certificate object"""

    __slots__ = ['publisher_id', 'work_digest', 'message_digest']

    def __init__(self, publisher_id, work_digest, message_digest):
        """
        publisher_id: fingerprint or Identity instance
        """
        if isinstance(publisher_id, Identity):
            self.publisher_id = publisher_id.fingerprint
        else:
            self.publisher_id = publisher_id
        self.work_digest = work_digest
        self.message_digest = message_digest

    def _get_xml(self):
        return ('certificate', [
            ('publisher_id', self.publisher_id),
            ('work_digest', self.work_digest),
            ('message_digest', self.message_digest),
        ])

    def get_reference(self): return GiftfileReference(self.message_digest)
    def get_id(self): return self.message_digest


def _sign_xml(xml_text, identity):
    return sign_xml(xml_text, identity.pgp_context, identity.fingerprint)


class ServerResponse(object):
    """Server response object

    Attributes:
        http_response - HttpResponse instance
    """

    __slots__ = ['http_response']

    def __init__(self, http_response):
        self.http_response = http_response


class LocationResponse(ServerResponse):
    """Server resource location response

    Attributes:
        http_response - HttpResponse instance
        url - URL of located resource
    """

    __slots__ = ['url']

    def __init__(self, http_response, url):
        ServerResponse.__init__(self, http_response)
        self.url = url


class ResourceResponse(ServerResponse):
    """Server resource response

    Attributes:
        http_response - HttpResponse instance
        libxml2_obj - the resource as a libxml2 document
        xml2tramp_obj - the resource via xml2tramp interface
        element_map - the resource via ElementMap

    Note that while libxml2 documents normally need to be explicitly
    freed after use, libxml2_obj is managed automatically and should not
    be freed.

    It is not the intention of this interface to impose a libxml2 or any
    other XML interface on the world.  This library manages decryption and
    signature verification of XML from the server, and as part of this service
    naturally must have already parsed the XML.  It makes sense then to make
    the parsed XML, in this implementation an libxml2 object, available to
    the user.

    If use of the libxml2 interface is unattractive, there are various
    wrappers which can provide a simpler interface.  This object provides
    two such wrappers via the xml2tramp_obj and element_map attributes.
    Note that the associated objects are only created if the attributes are
    accessed.  Finally, it is always possible to feed libxml2_obj.serialize()
    to the user's XML parser of choice.
    """

    __slots__ = ['libxml2_obj', '_xml2tramp_obj', '_element_map']

    def __init__(self, http_response, libxml2_obj):
        ServerResponse.__init__(self, http_response)
        self.libxml2_obj = libxml2_obj
        self._xml2tramp_obj = None
        self._element_map = None

    def _get_xml2tramp_obj(self):
        if not self._xml2tramp_obj:
            self._xml2tramp_obj = xml2tramp.CachingElement(self.libxml2_obj,
                default_namespace=XML_NAMESPACE)
        return self._xml2tramp_obj
    xml2tramp_obj = property(_get_xml2tramp_obj)

    def _get_element_map(self):
        if not self._element_map:
            self._element_map = ElementMap(self.xml2tramp_obj)
        return self._element_map
    element_map = property(_get_element_map)


def _get(resource_url, root_element_name, giftpool_id, entity_cache):
    """-> ResourceResponse instance

    root_element_name:  the expected name of the root element of the
        XML document.  If the name does not match, an AssertionError
        will be raised.

    giftpool_id:  Identity instance.  If not None, the server
        response will be checked for a proper signature, potentially raising
        SignatureVerificationFailed.

    entity_cache:  object implementing HTTP entity cache interface, or
        None to disable.  See MemoryHttpCache for an example.
    """
    response = http_open_xml(resource_url, entity_cache=entity_cache)
    response_entity = _check_http_response(response)
    doc = parse_memory(response_entity)
    if giftpool_id:
        _check_signature(doc, giftpool_id)
    root_element = doc.getRootElement()
    assert (XML_NAMESPACE, root_element_name) == (root_element.ns().content,
            root_element.name), \
        'root element is not "%s"' % root_element_name
    return ResourceResponse(response, doc)

def _request(request_url, xml_text):
    response = http_xml_request(request_url, str(xml_text))
    response_entity = _check_http_response(response)

def _request_new_resource(request_url, xml_text):
    """-> LocationResponse instance

    Use this when a request is expected to generate a new resource.
    """
    response = http_xml_request(request_url, str(xml_text))
    response_entity = _check_http_response(response, http_status.CREATED)
    url = response.getheader('Location', None)
    assert url is not None
    return LocationResponse(response, url)

def _locate(query_path, args):
    """-> LocationResponse instance"""
    query_url = '%s?%s' % (query_path, urlencode(args))
    response = http_open_xml(query_url)
    response_entity = _check_http_response(response, http_status.FOUND)
    url = response.getheader('Location', None)
    assert url is not None
    return LocationResponse(response, url)

def _create_donation(identity, issued_time):
    """-> XML text"""
    xml_data = [
        ('user_id',                 identity.fingerprint),
        ('issued',                  issued_time),
    ]
    xml_text = get_xml_element('donation_request', list_to_xml(xml_data),
        'xmlns="%s"' % XML_NAMESPACE)
    return _sign_xml(xml_text, identity)

def _create_allocation(identity, allocation_list, issued_time):
    """-> XML text

    allocation_list:
        list of (certificate, amount) or (giftfile_ref, amount) tuples
    """
    def get_allocation_xml((certificate_or_ref, amount)):
        return [
            certificate_or_ref._get_xml(),
            ('allocation_amount', amount),
        ]

    def get_allocation_list(name):
        result = []
        for allocation in allocation_list:
            result.append(('li', get_allocation_xml(allocation)))
        return result

    xml_data = [
        ('user_id',                 identity.fingerprint),
        ('allocations',             get_allocation_list),
        ('issued',                  issued_time),
    ]
    xml_text = get_xml_element('allocation_request', list_to_xml(xml_data),
        'xmlns="%s"' % XML_NAMESPACE)
    return _sign_xml(xml_text, identity)

def _create_remittance(identity, amount, payee_name, delivery_address,
        issued_time):
    """-> XML text"""
    xml_data = [
        ('user_id',                 identity.fingerprint),
        ('remittance_amount',       amount),
        ('payee_name',              payee_name),
        ('delivery_address',        delivery_address),
        ('issued',                  issued_time),
    ]
    xml_text = get_xml_element('remittance_request', list_to_xml(xml_data),
        'xmlns="%s"' % XML_NAMESPACE)
    return _sign_xml(xml_text, identity)

def _create_record_deletion(identity, record_list, issued_time):
    """-> XML text"""

    def get_record_list(name):
        result = []
        for record in record_list:
            result.append(('li', record))
        return result

    xml_data = [
        ('user_id',                 identity.fingerprint),
        ('records',                 get_record_list),
        ('issued',                  issued_time),
    ]
    xml_text = get_xml_element('record_deletion_request',
        list_to_xml(xml_data), 'xmlns="%s"' % XML_NAMESPACE)
    return _sign_xml(xml_text, identity)

def _create_donation_update(identity, target_uri, issued_time,
        is_received=False, amount=None, payer_name=None, check_number=None,
        is_cleared=False, is_denied=False, reason=None):
    """-> XML text"""

    def get_donation_received_xml():
        return [
            ('donation_amount',         amount),
            ('payer_name',              payer_name),
            ('payment_type',            'check'),
            ('check_number',            check_number),
        ]

    xml_data = [
        ('user_id',                 identity.fingerprint),
        ('donation',                target_uri),
        ('issued',                  issued_time),
    ]
    if is_received:
        assert amount and payer_name and check_number
        xml_data.extend([
            ('donation_received', get_donation_received_xml()),
        ])
    if is_cleared:
        xml_data.extend([
            ('donation_cleared', None),
        ])
    if is_denied:
        assert reason
        xml_data.extend([
            ('donation_denied', [('reason', reason)]),
        ])
    xml_text = get_xml_element('donation_update', list_to_xml(xml_data),
        'xmlns="%s"' % XML_NAMESPACE)
    return _sign_xml(xml_text, identity)

def _create_remittance_update(identity, target_uri, issued_time,
        is_issued=False, check_number=None, check_date=None,
        is_reconciled=False, is_denied=False, reason=None):
    """-> XML text"""
    xml_data = [
        ('user_id',                 identity.fingerprint),
        ('remittance',              target_uri),
        ('issued',                  issued_time),
    ]
    if is_issued:
        assert check_number and check_date
        xml_data.extend([
            ('remittance_issued', [
                ('check_number', check_number),
                ('check_date', check_date),
            ]),
        ])
    if is_reconciled:
        xml_data.extend([
            ('remittance_reconciled', None),
        ])
    if is_denied:
        assert reason
        xml_data.extend([
            ('remittance_denied', [('reason', reason)]),
        ])

    xml_text = get_xml_element('remittance_update', list_to_xml(xml_data),
        'xmlns="%s"' % XML_NAMESPACE)
    return _sign_xml(xml_text, identity)

def _create_jobs_update(identity, issued_time):
    """-> XML text"""
    xml_data = [
        ('user_id',                 identity.fingerprint),
        ('issued',                  issued_time),
    ]
    xml_text = get_xml_element('jobs_update', list_to_xml(xml_data),
            'xmlns="%s"' % XML_NAMESPACE)
    return _sign_xml(xml_text, identity)


class TransactionRequest(object):
    """
    Server transaction request

    Attributes:
        url
        type
        xml_text
    """

    _NEW_RESOURCE_REQUESTS = [
        'donation_request',
        'allocation_request',
        'remittance_request',
    ]

    def __init__(self, url, type, xml_text):
        self.url = url
        self.type = type
        self.xml_text = xml_text

    def send(self):
        """-> LocationResponse (if new resource created) or None"""
        if self.type in self._NEW_RESOURCE_REQUESTS:
            return _request_new_resource(self.url, self.xml_text)
        else:
            _request(self.url, self.xml_text)


def _filter_time(x):
    if x is None:
        return get_date_and_time()
    else:
        return x


class GiftpoolInterface(object):
    """Giftpool interface

    This class represents an interface to a specific giftpool, through which
    requests and queries may be made.

    The following URL variables are available as attributes:
        info
        private_record_search
        public_record_search
        donation_request
        allocation_request
        remittance_request
        record_deletion_request
        donation_update
        remittance_update
        jobs_update
    """

    _interface_items = [
        'info',
        'private_record_search',
        'public_record_search',
        'donation_request',
        'allocation_request',
        'remittance_request',
        'record_deletion_request',
    ]

    _admin_interface_items = [
        'donation_update',
        'remittance_update',
        'jobs_update',
    ]

    def __init__(self, giftpool_url, giftpool_id, entity_cache=None):
        """
        giftpool_id should be an Identity instance.  It may also be an
        AutoIdentity instance or None, but please be aware of the security
        issues these raise.  If None, then the signature of server responses
        will not be verified.  If an AutoIdentity instance, the ID used to
        sign the giftpool interface resource (retrieved from the server) will
        be considered authoritative and will be used for subsequent response
        verification.  This assumes that the giftpool's public key is
        present in the keyring (SignatureVerificationFailed will be raised
        otherwise).

        entity_cache: to enable transparent caching of HTTP response
            entities, set this to any object providing the HTTP entity cache
            interface.  See MemoryHttpCache for an example.  Note that the
            caching method used will never return a stale result.

        The create_ methods return a TransactionRequest object.  It is
        intended that the user save this object to reliable storage
        before sending the request, in support of reliable messaging.

        TODO: add user identity to signature of get_* methods
            This will be necessary once private record encryption is
            implemented.
        """
        self.giftpool_id = giftpool_id
        self.entity_cache = entity_cache

        # set interface vars
        interfaces = _get(giftpool_url, 'interfaces', giftpool_id,
            entity_cache).element_map
        interface = interfaces['interface']
        for item in self._interface_items:
            setattr(self, item, interface[item])
        admin_interface = interface['admin']
        for item in self._admin_interface_items:
            setattr(self, item, admin_interface[item])

    def _get(self, url, resource_name):
        if resource_name == 'info':
            record_name = 'giftpool_record'
        else:
            record_name = resource_name + '_record'
        return _get(url, record_name, self.giftpool_id, self.entity_cache)

    def get_info(self):
        """-> ResourceResponse instance"""
        return self._get(self.info, 'info')

    def get_user(self, url):
        """-> ResourceResponse instance"""
        return self._get(url, 'user')

    def get_giftfile(self, url):
        """-> ResourceResponse instance"""
        return self._get(url, 'giftfile')

    def get_donation(self, url):
        """-> ResourceResponse instance"""
        return self._get(url, 'donation')

    def get_allocation(self, url):
        """-> ResourceResponse instance"""
        return self._get(url, 'allocation')

    def get_remittance(self, url):
        """-> ResourceResponse instance"""
        return self._get(url, 'remittance')

    def locate_user_record(self, user_fingerprint, public=False):
        """-> LocationResponse instance"""
        if public: search = self.public_record_search
        else: search = self.private_record_search
        return _locate(search, {'user' : user_fingerprint})

    def locate_giftfile_record(self, giftfile_digest, public=False):
        """-> LocationResponse instance"""
        if public: search = self.public_record_search
        else: search = self.private_record_search
        return _locate(search, {'giftfile' : giftfile_digest})

    def create_donation(self, identity, issued_time=None):
        """-> TransactionRequest instance"""
        issued_time = _filter_time(issued_time)
        xml_text = _create_donation(identity, issued_time)
        return TransactionRequest(
            self.donation_request,
            'donation_request',
            xml_text)

    def create_allocation(self, identity, allocation_list, issued_time=None):
        """-> TransactionRequest instance

        allocation_list:
            list of (certificate, amount) or (giftfile_ref, amount) tuples
        """
        issued_time = _filter_time(issued_time)
        xml_text = _create_allocation(identity, allocation_list,
            issued_time)
        return TransactionRequest(
            self.allocation_request,
            'allocation_request',
            xml_text)

    def create_remittance(self, identity, amount, payee_name,
            delivery_address, issued_time=None):
        """-> TransactionRequest instance"""
        issued_time = _filter_time(issued_time)
        xml_text = _create_remittance(identity, amount, payee_name,
            delivery_address, issued_time)
        return TransactionRequest(
            self.remittance_request,
            'remittance_request',
            xml_text)

    def create_record_deletion(self, identity, record_list, issued_time=None):
        """-> TransactionRequest instance"""
        issued_time = _filter_time(issued_time)
        xml_text = _create_record_deletion(identity, record_list,
            issued_time)
        return TransactionRequest(
            self.record_deletion_request,
            'record_deletion_request',
            xml_text)

    def create_donation_update(self, identity, target_uri, issued_time=None,
            is_received=False, amount=None, payer_name=None, check_number=None,
            is_cleared=False, is_denied=False, reason=None):
        """-> TransactionRequest instance"""
        issued_time = _filter_time(issued_time)
        xml_text = _create_donation_update(identity, target_uri,
            issued_time, is_received, amount, payer_name, check_number,
            is_cleared, is_denied, reason)
        return TransactionRequest(
            self.donation_update,
            'donation_update',
            xml_text)

    def create_remittance_update(self, identity, target_uri, issued_time=None,
            is_issued=False, check_number=None, check_date=None,
            is_reconciled=False, is_denied=False, reason=None):
        """-> TransactionRequest instance"""
        issued_time = _filter_time(issued_time)
        xml_text = _create_remittance_update(identity, target_uri, issued_time,
            is_issued, check_number, check_date, is_reconciled, is_denied,
            reason)
        return TransactionRequest(
            self.remittance_update,
            'remittance_update',
            xml_text)

    def create_jobs_update(self, identity, issued_time=None):
        """-> TransactionRequest instance"""
        issued_time = _filter_time(issued_time)
        xml_text = _create_jobs_update(identity, issued_time)
        return TransactionRequest(
            self.jobs_update,
            'jobs_update',
            xml_text)
