# Copyright (C) 2004  John Belmonte

"""Giftpool interface with sugar

Contained here is a replacement for clientlib.GiftpoolInterface which
provides an interface more native to Python.
"""
# TODO: complete native interfaces to resources, document record attributes

from types import StringTypes
from giftfile import clientlib, transaction_status
from giftfile.fixedpoint import FixedPoint as Decimal
from giftfile.time_util import get_seconds


def _init_base(obj, *args):
    """Initialize base of class"""
    obj.__class__.__base__.__init__(obj, *args)


class _Record(object):
    """
    This object wraps an ElementMap, providing access to the content
    of root elements as Python attributes.  It also can perform type
    conversion of the content.

    Currently, type conversion is automatic and should be regarded as
    a hack.  It should be replaced with something that knows the
    actual type of each field.
    """

    __slots__ = ['_element_map']

    def __init__(self, element_map):
        self._element_map = element_map

    def _convert(self, name, content):
        if not isinstance(content, StringTypes):
            return content
        # don't convert ID's to a number
        if name.endswith('id') or name in ['check_number']:
            return content
        if content.isdigit(): return int(content)
        try: return Decimal(content)
        except ValueError: pass
        try: return get_seconds(content)
        except ValueError: pass
        return content

    def __getattr__(self, name):
        """-> content of root-level record field"""
        return self._convert(name, self._element_map[name])


class _Transaction(object):
    """Mixin in deriving status_code attribute from status"""
    __slots__ = []
    def __get_status_code(self):
        return transaction_status.get_code(self.status)
    status_code = property(__get_status_code)


class GiftpoolRecord(_Record):
    __slots__ = ['http_response', 'user_testing_capability']

    def __init__(self, resource_response):
        self.http_response = resource_response.http_response
        _init_base(self, resource_response.element_map)

        self.user_testing_capability = 'user_testing_capability' in \
            resource_response.element_map


class DonationRecord(_Record, _Transaction):
    __slots__ = ['http_response']

    def __init__(self, resource_response):
        self.http_response = resource_response.http_response
        _init_base(self, resource_response.element_map)


class UserRecord(_Record, _Transaction):
    __slots__ = ['http_response', 'donations', 'allocations', 'giftfiles',
        'remittances']

    def __init__(self, resource_response):
        self.http_response = resource_response.http_response
        _init_base(self, resource_response.element_map)
        # ISSUE: if these lists are very large, and don't end up being
        #   used, we may be wasting resources.
        for url_list in ['donations', 'allocations', 'giftfiles',
                'remittances']:
            list_value = self.__getattr__(url_list)
            if not list_value:
                result = []
            else:
                result = [item for item in list_value]
            setattr(self, url_list, result)


class SuballocationRecord(_Record, _Transaction):
    """_Record for an individual allocation"""
    __slots__ = []


class AllocationRecord(_Record, _Transaction):
    __slots__ = ['http_response']

    def __init__(self, resource_response):
        self.http_response = resource_response.http_response
        _init_base(self, resource_response.element_map)

    def _get_allocations(self):
        """-> list of SuballocationRecord instances"""
        return map(SuballocationRecord, self.__getattr__('allocations'))
    allocations = property(_get_allocations)


class GiftfileRecord(_Record):
    __slots__ = ['http_response']

    def __init__(self, resource_response):
        self.http_response = resource_response.http_response
        _init_base(self, resource_response.element_map)


class RemittanceRecord(_Record, _Transaction):
    __slots__ = ['http_response']

    def __init__(self, resource_response):
        self.http_response = resource_response.http_response
        _init_base(self, resource_response.element_map)


class GiftpoolInterface(clientlib.GiftpoolInterface):

    def get_info(self):
        """-> GiftpoolRecord instance"""
        return GiftpoolRecord(clientlib.GiftpoolInterface.get_info(self))

    def get_user(self, url):
        """-> UserRecord instance"""
        return UserRecord(clientlib.GiftpoolInterface.get_user(self, url))

    def get_giftfile(self, url):
        """-> GiftfileRecord instance"""
        return GiftfileRecord(
            clientlib.GiftpoolInterface.get_giftfile(self, url))

    def get_donation(self, url):
        """-> DonationRecord instance"""
        return DonationRecord(
            clientlib.GiftpoolInterface.get_donation(self, url))

    def get_allocation(self, url):
        """-> AllocationRecord instance"""
        return AllocationRecord(
            clientlib.GiftpoolInterface.get_allocation(self, url))

    def get_remittance(self, url):
        """-> RemittanceRecord instance"""
        return RemittanceRecord(
            clientlib.GiftpoolInterface.get_remittance(self, url))


def _test():
    from giftfile import clientlib
    pgp_homedir='~/var/giftpool-client/pgp'
    giftpool_url = 'http://john:2080/giftpool/'
    giftpool_id = None
    giftpool = GiftpoolInterface(giftpool_url, giftpool_id)
    giftpool_info = giftpool.get_info()
    print giftpool_info.name
    allocation_url = 'http://john:2080/giftpool/1/private/allocations/' \
        'e54a512251ea4d61b5f4a36ebb84c3ee74991c71'
    allocation_rec = giftpool.get_allocation(allocation_url)
    print allocation_rec.status_code
    print allocation_rec.allocations[0].status_code

if __name__ == '__main__':
    _test()
