# Copyright (C) 2004  Anthony Carrico

"""A functional xml library.

Features (or misfeatures, ymmv):
  * Xml 1.0 with the xml namespace constraints
    * http://www.w3.org/TR/2004/REC-xml-20040204/
    * http://www.w3.org/TR/REC-xml-names/
    * http://www.jclark.com/xml/xmlns.htm
    * http://www.w3.org/2000/09/xppa
  * Nodes are immutable. An operation doesn't actually change a node,
    but rather returns a new node (that may share structure with the
    original).
  * Nodes can exist with parents ("adopted") or without
    ("orphaned"). Of course, root nodes are always orphans. We
    follow xpath's slightly confusing terminology, so while an
    adopted node has a parent, it isn't necessarily a child. In
    particular only element, text, processing instruction, and
    comment nodes are potentially children. Attribute, namespace
    declaration, and of course root nodes are never children.
  * Nodes can share structure. You can create a node and use it in
    two documents. You can extract a node from one document and use
    it in another document without destroying the original.
  * It is efficient to analyze a document from the top down or the
    bottom up.
  * It is most efficient to synthesize a document from the bottom
    up. This is because changing an adopted node copies its
    ancestors. However, this is not terribly inefficient since many
    documents are shallow, and because siblings, aunts, cousins,
    etc. are shared.
  * Element types are unambiguous outside their parent context.
  * Global attribute names are unambiguous outside their parent
    context.
  * Per element attribute names are ambiguous outside their parent
    context. See http://www.w3.org/TR/REC-xml-names/#ns-expnames
  * libxml2 export
  * libxml2 import
  * Parsing (via libxml2)
  * Serialization (via libxml2)
  * RelaxNG schema support (via libxml2)
  * C14N support (via libxml2)

Examples:
  * Many examples are at the bottom of this file

Definitions:
  * NCName -- no colon name, see http://www.w3.org/TR/REC-xml-names/#NT-NCName
  * content -- an element's children, see http://www.w3.org/TR/REC-xml/#dt-content

Current limitations:
  * Doesn't check if NCName, uri, attribute-value, (other?) are legal.
  * etc...

Maybe:
  * Signature
  * Encryption

Not currently planned:
  * entity support
  * dtd support (dtd doesn't support namespaces)
  * xml schema support (except datatypes, for RelaxNG)

Minimum requirements:
  * python 2.3
  * libxml 2.6.6 (or so)
    * libxml 2.6.3+ ok, except for NsDecl serialization bug.

Architecture (for the curious):

  A typical imperative approach to modeling XPath elements is double
  linked:

  element node:
    attribute node set ref <-> attribute nodes
    namespace declaration refs <-> namespace declaration nodes
    content refs <-> child nodes
    parent node ref <-> parent node.

  This neglects the namespace node set axis since it is implict in the
  element's and its ancestors' local namespace declarations.

  In contrast, the funkxml model is single linked, which enables the
  data structure to behave functionally. The abstract funkxml node can
  actually be one of two concrete types: orphaned or adopted. A
  concrete adopted node is primarily an envelope for a concrete
  orphaned node which serves as the representation and is shared with
  the parent:

  adopted element node:
    parent ref -> parent node:
                    child ref -\\
    orphan ref -----------------+-> orphaned element node:
                                      attribute node set ref -> ...
                                      namespace declaration refs -> ...
                                      content refs -> ...

"""

from giftfile import srfi1
from giftfile import tuple_util
import libxml2

class Error(Exception):
    def __init__(self, message=''):
        Exception.__init__(self, message)

class MissingNsDeclError(Error):
    def __init__(self):
        Error.__init__(self)

class EmptyTextError(Error):
    def __init__(self):
        Error.__init__(self)

class MissingRootElementError(Error):
    def __init__(self):
        Error.__init__(self)

class MultipleRootElementError(Error):
    def __init__(self):
        Error.__init__(self)

class NameError(Error):
    """Raised when the name of a node violates some xml rule.

    A PI namespace must be None.

    A global attribute name must have a prefix."""

    def __init__(self):
        Error.__init__(self)

class PrefixClashError(Error):
    def __init__(self):
        Error.__init__(self)

class PrefixReservedError(Error):
    def __init__(self):
        Error.__init__(self)

class NsDeclClashError(Error):
    def __init__(self):
        Error.__init__(self)

class AttributeClashError(Error):
    def __init__(self):
        Error.__init__(self)

class ParseFileError(Error):
    def __init__(self, filename):
        msg = 'Parse of file %s failed' % repr(filename)
        Error.__init__(self, msg)

class QName(object):
    """Qualified name."""

    __slots__ = ['_prefix', '_local']
    prefix = property(fget=lambda self: self._prefix,
                      doc="None if the QName has no prefix.")
    local = property(fget=lambda self: self._local)

    def __init__(self, prefix, local):
        self._prefix = prefix
        self._local = local

    def __str__(self):
        if prefix:
            return prefix + ':' + local
        else:
            return local

class ExpName(object):
    """The expanded name of an element or an attribute.

    Use an ns of None for no namespace. An attribute name with an ns
    uri is a global attribute name. An attribute name with an ns of
    None is a per-element attribute name.

    NOTE: ExpNames are cached internally, so there is only ever one
    instance of a particular expanded name.

    NOTE: "The use of relative URI references, including same-document
    references, in namespace declarations is deprecated. Future W3C
    specifications will define no interpretation for them."
    """

    _cache = {}

    def _check_local(local):
        if ((not isinstance(local, (str, unicode)))
            or (local == '')):
            raise TypeError
    _check_local=staticmethod(_check_local)

    def _check_ns(ns):
        if ((not isinstance(ns, (str, unicode, type(None))))
            or (ns == '')):
            raise TypeError
    _check_ns=staticmethod(_check_ns)

    __slots__ = ['_local', '_ns', '__cmpstr']
    local = property(lambda self: self._local,
                    doc="an NCName")
    ns = property(lambda self: self._ns,
                  doc="a namespace uri, or None")

    def __new__(cls, local, ns=None):
        ExpName._check_local(local)
        ExpName._check_ns(ns)
        key = (local, ns)
        try:
            return ExpName._cache[key]
        except KeyError:
            name = object.__new__(cls)
            name._local = local
            name._ns = ns
            name.__cmpstr = (ns or '') + ':' + local
            ExpName._cache[key]=name
            return name

    def __init__(self, local, ns=None):
        pass

    def __lt__(self, other):
        return self.__cmpstr.__lt__(other.__cmpstr)
    def __le__(self, other):
        return self.__cmpstr.__le__(other.__cmpstr)
    def __eq__(self, other):
        return self.__cmpstr.__eq__(other.__cmpstr)
    def __ne__(self, other):
        return self.__cmpstr.__ne__(other.__cmpstr)
    def __gt__(self, other):
        return self.__cmpstr.__gt__(other.__cmpstr)
    def __ge__(self, other):
        return self.__cmpstr.__ge__(other.__cmpstr)
    def __cmp__(self, other):
        return self.__cmpstr.__cmp__(other.__cmpstr)
    def __hash__(self):
        return self.__cmpstr.__hash__()

    def __str__(self):
        if self.ns:
            return '{' + self.ns + '}:' + self.local
        else:
            return self.local

_xml_ns = 'http://www.w3.org/XML/1998/namespace'

_default_prefix_map = {"xml":_xml_ns, None:None}

class _FreePrefixMap(dict):
    __slots__= ['bound_ns_decls']

    def __init__(self, bound_ns_decls=None):
        dict.__init__(self)
        if not ((bound_ns_decls == None)
                or isinstance(bound_ns_decls, NsDeclMap)):
            raise TypeError
        self.bound_ns_decls = bound_ns_decls

    def check_default(self):
        for prefix in self:
            if ((prefix not in _default_prefix_map)
                or _default_prefix_map[prefix] != self[prefix]):
                raise MissingNsDeclError

    def ref_prefix(self, prefix, ns):
        if ((self.bound_ns_decls != None)
            and (prefix in self.bound_ns_decls)):
            if self.bound_ns_decls[prefix].ns != ns:
                raise PrefixClashError
        elif prefix in self:
            if self[prefix] != ns:
                raise PrefixClashError
        else:
            self[prefix] = ns

    def ref_attribute_prefix(self, prefix, ns):
        # NOTE: An attribute name prefix of None identifies a
        # per-element attribute name, not a global name with a
        # default namespace.
        if prefix != None:
            ref_prefix(prefix, ns)

    def merge(self, other):
        for prefix in other:
            self.ref_prefix(prefix, other[prefix])

    def get_closed_ns_decls(self):
        if self.bound_ns_decls == None:
            result = []
        else:
            result = self.bound_ns_decls.values()
        for prefix, ns in self.iteritems():
            result.append(NsDecl(prefix, ns))
        return NsDeclMap(result)

    def get_conflicting_ns_decls(self, free_prefix_map):
        nsds = []
        for prefix, ns in free_prefix_map.iteritems():
            if ((self.bound_ns_decls != None)
                and (prefix in self.bound_ns_decls)):
                if self.bound_ns_decls[prefix].ns != ns:
                    nsds.append(NsDecl(prefix, ns))
            elif prefix in self:
                if self[prefix] != ns:
                    nsds.append(NsDecl(prefix, ns))
        return nsds

def _resolve_prefix(prefix, env):
    """env is a srfi1 association list from prefix -> libxml2.xmlNs"""
    binding = srfi1.assoc(prefix, env)
    if binding:
        return binding[1]
    elif prefix == 'xml':
        return None
    elif prefix == None:
        return None
    else:
        raise MissingNsDeclError

def _check_string_value(string_value):
    if not isinstance(string_value, (str, unicode)):
        raise TypeError
    else:
        return string_value

def _check_name(prefix, name):
    if isinstance(name, (str, unicode)):
        name = ExpName(name)
    elif not isinstance(name, ExpName):
        raise TypeError
    if isinstance(prefix, (str, unicode)):
        if (prefix == 'xml') and (name.ns != _xml_Ns):
            raise PrefixReservedError
    elif prefix!=None:
        raise TypeError
    return prefix, name

class _ImmutableMap(object):
    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, name):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError

    def __contains__(self, name):
        raise NotImplementedError

    def keys(self):
        raise NotImplementedError

    def values(self):
        raise NotImplementedError

    def items(self):
        raise NotImplementedError

    def has_key(self, key):
        raise NotImplementedError

    def get(self, name, x):
        raise NotImplementedError

    def iterkeys(self):
        raise NotImplementedError

    def itervalues(self):
        raise NotImplementedError

    def iteritems(self):
        raise NotImplementedError

    def __eq__(self, other):
        raise NotImplementedError

    def __ne__(self, other):
        return not self.__eq__(other)

class NsDeclMap(_ImmutableMap):
    """A mapping from prefix to NsDecl."""

    __slots__ = []

    parent = property(doc="the NsDecls' parent as defined in XPath or None.")
    orphan = property(doc="the NsDeclMap without a parent.")

    def __new__(cls, *rest, **krest):
        if (cls == NsDeclMap): cls = _OrphanedNsDeclMap
        return object.__new__(cls, *rest, **krest)

    def __init__(self, ns_decls=()):
        raise NotImplementedError

    def is_orphan(self):
        raise NotImplementedError

class _OrphanedNsDeclMap(NsDeclMap):
    """An implementation of NsDeclMap without a parent."""

    __slots__ = ['_map']

    parent = property(lambda self: None)
    orphan = property(lambda self: self)

    def __init__(self, ns_decls=()):
        ns_decl_map = {}
        for ns_decl in ns_decls:
            if not isinstance(ns_decl, NsDecl):
                raise TypeError
            ns_decl = ns_decl.orphan
            if ns_decl.prefix in ns_decl_map:
                raise NsDeclClashError
            ns_decl_map[ns_decl.prefix]=ns_decl
        self._map = ns_decl_map

    def is_orphan(self): return True

    def _set_parent(self, parent):
        return _AdoptedNsDeclMap(self, parent)

    def __len__(self):
        return self._map.__len__()

    def __getitem__(self, name):
        return self._map[name]

    def __iter__(self):
        return self.iterkeys()

    def __contains__(self, name):
        return self._map.__contains__(name)

    def keys(self):
        return self._map.keys()

    def values(self):
        return self._map.values()

    def items(self):
        return self._map.items()

    def has_key(self, name):
        return self._map.has_key()

    def get(self, name, x):
        return self._map.get(name, x)

    def iterkeys(self):
        return self._map.iterkeys()

    def itervalues(self):
        return self._map.itervalues()

    def iteritems(self):
        return self._map.iteritems()

    def __eq__(self, other):
        return ((id(self) == id(other))
                or (isinstance(other, _OrphanedNsDeclMap)
                    and (self._map == other._map)))

class _AdoptedNsDeclMap(NsDeclMap):
    """An implementation of NsDeclMap with a parent."""

    __slots__ = ['_orphan', '_parent']
    parent = property(lambda self: self._parent)
    orphan = property(lambda self: self._orphan)

    def __init__(self, orphan, parent):
        self._orphan = orphan
        self._parent = parent

    def is_orphan(self): return False

    def _set_parent(self, parent):
        return _AdoptedNsDeclMap(self._orphan, parent)

    def __len__(self):
        return self._orphan.__len__()

    def __getitem__(self, key):
        return self._orphan.__getitem__(
            key)._set_parent(self._parent)

    def __iter__(self):
        return self.iterkeys()

    def __contains__(self, key):
        return self._orphan.__contains__(key)

    def keys(self):
        return self._orphan.keys()

    def values(self):
        return self._orphan.values()

    def items(self):
        return self._orphan.items()

    def has_key(self, name):
        return self._orphan.has_key()

    def get(self, key, x):
        value = self._orphan.get(key, None)
        if value == None:
            return x
        else:
            return value._set_parent(self._parent)

    def iterkeys(self):
        return self._orphan.iterkeys()

    def itervalues(self):
        for value in self._orphan.itervalues():
            yield value._set_parent(self._parent)

    def iteritems(self):
        for name, value in self._orphan.iteritems():
            yield (name, value._set_parent(self._parent))

    def __eq__(self, other):
        return ((id(self) == id(other))
                or (isinstance(other, _AdoptedNsDeclMap)
                    and (self._parent == other._parent)))

class AttrMap(_ImmutableMap):
    """A mapping from ExpName to Attribute."""

    __slots__ = []

    parent = property(doc="the attributes' parent as defined in XPath or None.")
    orphan = property(doc="the attribute map without a parent.")

    def __new__(cls, *rest, **krest):
        if (cls == AttrMap): cls = _OrphanedAttrMap
        return object.__new__(cls, *rest, **krest)

    def __init__(self, attributes=()):
        raise NotImplementedError

    def is_orphan(self):
        raise NotImplementedError

class _OrphanedAttrMap(AttrMap):
    """A mapping from ExpName to Attribute."""

    __slots__ = ['_attr_map', '_free_prefix_map']

    parent = property(lambda self: None)
    orphan = property(lambda self: self)

    def __init__(self, attributes=()):
        free_prefix_map = _FreePrefixMap()
        attr_map = {}

        for attr in attributes:
            if not isinstance(attr, Attribute):
                raise TypeError
            attr = attr.orphan
            if attr.name in attr_map:
                raise AttributeClashError
            free_prefix_map.ref_attribute_prefix(
                attr.prefix, attr.name.ns)
            attr_map[attr.name]=attr

        self._attr_map = attr_map
        self._free_prefix_map = free_prefix_map

    def is_orphan(self): return True

    def _set_parent(self, parent):
        return _AdoptedAttrMap(self, parent)

    def __len__(self):
        return self._attr_map.__len__()

    def __getitem__(self, name):
        if isinstance(name, (str, unicode)):
            name = ExpName(name)
        return self._attr_map[name]

    def __iter__(self):
        return self.iterkeys()

    def __contains__(self, name):
        return self._attr_map.__contains__(name)

    def keys(self):
        return self._attr_map.keys()

    def values(self):
        return self._attr_map.values()

    def items(self):
        return self._attr_map.items()

    def has_key(self, name):
        if isinstance(name, (str, unicode)):
            name = ExpName(name)
        return self._attr_map.has_key()

    def get(self, name, x):
        if isinstance(name, (str, unicode)):
            name = ExpName(name)
        return self._attr_map.get(name, x)

    def iterkeys(self):
        return self._attr_map.iterkeys()

    def itervalues(self):
        return self._attr_map.itervalues()

    def iteritems(self):
        return self._attr_map.iteritems()

    def __eq__(self, other):
        return ((id(self) == id(other))
                or (isinstance(other, _OrphanedAttrMap)
                    and (self._attr_map == other._attr_map)))

class _AdoptedAttrMap(AttrMap):
    """A mapping from ExpName to Attribute."""

    __slots__ = ['_orphan', '_parent']
    parent = property(lambda self: self._parent)
    orphan = property(lambda self: self._orphan)

    def __init__(self, orphan, parent):
        self._orphan = orphan
        self._parent = parent

    def is_orphan(self): return False

    def _set_parent(self, parent):
        return _AdoptedAttrMap(self._orphan, parent)

    def __len__(self):
        return self._orphan.__len__()

    def __getitem__(self, name):
        return self._orphan.__getitem__(
            name)._set_parent(self._parent)

    def __iter__(self):
        return self.iterkeys()

    def __contains__(self, name):
        return self._orphan.__contains__(name)

    def keys(self):
        return self._orphan.keys()

    def values(self):
        return self._orphan.values()

    def items(self):
        return self._orphan.items()

    def has_key(self, name):
        return self._orphan.has_key()

    def get(self, name, x):
        value = self._orphan.get(key, None)
        if value == None:
            return x
        else:
            return value._set_parent(self._parent)

    def iterkeys(self):
        return self._orphan.iterkeys()

    def itervalues(self):
        for attr in self._orphan.itervalues():
            yield attr._set_parent(self._parent)

    def iteritems(self):
        for name, attr in self._orphan.iteritems():
            yield (name, attr._set_parent(self._parent))

    def __eq__(self, other):
        return ((id(self) == id(other))
                or (isinstance(other, _AdoptedAttrMap)
                    and (self._parent == other._parent)))

class ContentMap(object):
    """A mapping from ExpName to sequence of Element over Content.

    The elements that constitute each value are in document order."""

    __slots__ = ['_content']

    parent = property(lambda self: self._content.parent,
                      doc="The mapped Content's parent.")

    def __get_orphan(self):
        if self.is_orphan(): return self
        else: return self._content.orphan.content_map

    orphan = property(__get_orphan,
                      doc="The ContentMap of the mapped Content's orphan.")

    def __init__(self, content):
        if not isinstance(content, Content):
            raise TypeError
        else:
            self._content = content

    def is_orphan(self):
        return self._content.is_orphan()

    def __len__(self):
        return self._content.content_map.__len__()

    def __indexes_to_elements(self, indexes):
        return tuple(map(lambda index: self._content.__getitem__(index),
                         indexes))

    def __getitem__(self, name):
        if isinstance(name, (str, unicode)):
            name = ExpName(name)
        indexes = self._content.orphan._name_map[name]
        return self.__indexes_to_elements(indexes)

    def __iter__(self):
        return self.iterkeys()

    def __contains__(self, name):
        if isinstance(name, (str, unicode)):
            name = ExpName(name)
        return self._content.orphan._name_map.__contains__(name)

    def keys(self):
        return self._content.orphan._name_map.keys()

    def values(self):
        raise NotImplementedError

    def items(self):
        raise NotImplementedError

    def has_key(self, key):
        return self._content.orphan._name_map.has_key()

    def get(self, name, x):
        if isinstance(name, (str, unicode)):
            name = ExpName(name)
        indexes = self._content.orphan._name_map.get(name, ())
        if len(indexes) == 0:
            return x
        else:
            return self.__indexes_to_elements(indexes)

    def iterkeys(self):
        return self._content.orphan._name_map.iterkeys()

    def itervalues(self):
        content = self._content
        for name, indexes in content.orphan._name_map.iteritems():
            yield self.__indexes_to_elements(indexes)

    def iteritems(self):
        content = self._content
        for name, indexes in content.orphan._name_map.iteritems():
            yield (name, self.__indexes_to_elements(indexes))

class Content(object):
    """A sequence of Element, Text, Comment, and/or PI nodes.

    If the Content was extracted from a parent node (Root or
    Element), the Content has a parent. Like a Node, a newly
    constructed Content is an orphan."""

    __slots__ = []

    parent = property(doc="The content's parent as defined in XPath or None.")
    orphan = property(doc="The Content without a parent.")
    string_value = property(doc="The string value as defined for element nodes in XPath.")
    content_map = property(lambda self: ContentMap(self),
                         doc="A ContentMap of the Content's elements.")

    def __new__(cls, *rest, **krest):
        if (cls == Content): cls = _OrphanedContent
        return object.__new__(cls, *rest, **krest)

    def __init__(self, children=()):
        raise NotImplementedError

    def is_orphan(self):
        raise NotImplementedError

    def with_child(self, child, index):
        """Return a version of this node with different child"""
        raise NotImplementedError

    def __len__(self): raise NotImplementedError
    def __getitem__(self, key): raise NotImplementedError
    def __iter__(self): raise NotImplementedError
    def __contains__(self, item): raise NotImplementedError

    def __add__(self, other):
        # ISSUE: This could be more efficient
        x = self.orphan._children
        if isinstance(other, Content):
            return Content(x + other.orphan._children)
        elif isinstance(other, tuple):
            return Content(x + other)
        else:
            return Content(x + Content(other))

    def __mul__(self, other):
        # ISSUE: This could be more efficient
        return Content(self.orphan._children * other)

class _OrphanedContent(Content):
    """Implementation of Content without a parent."""

    __slots__ = ['_children', '_name_map', '_free_prefix_map']

    parent = property(lambda self: None)
    orphan = property(lambda self: self)

    def __get_string_value(self):
        result = ''
        for ch in self._children:
            if isinstance(ch, (Text, Element)):
                result += ch.string_value
        return result

    string_value = property(__get_string_value)

    def __init__(self, children=()):
        name_map = {}
        free_prefix_map = _FreePrefixMap()

        def process_element(element):
            free_prefix_map.merge(element._free_prefix_map)
            try:
                name_map[element.name].append(index)
            except KeyError:
                name_map[element.name] = [index]

        # Check if the children are normal and process elements:
        text = False
        index = 0
        if isinstance(children, tuple):
            while index < len(children):
                child = children[index]
                if isinstance(child, Text):
                    if text or not child.is_orphan(): break
                    else: text = True
                elif isinstance(child, (Comment, PI)):
                    if not child.is_orphan(): break
                    text = False
                elif isinstance(child, Element):
                    if not child.is_orphan(): break
                    text = False
                    process_element(child)
                else:
                    raise TypeError
                index += 1

        # If the children aren't normal, then normalize them, and
        # process the rest of the elements:
        if not isinstance(children, tuple) or (index < len(children)):
            normal_children = list(children[0:index])
            while index < len(children):
                child = children[index]
                if isinstance(child, Text):
                    child = child.orphan
                    if text:
                        normal_children[-1] = Text(
                            normal_children[-1].string_value
                            + child.string_value)
                    else:
                        text = True
                        normal_children.append(child)
                elif isinstance(child, (Comment, PI)):
                    text = False
                    child = child.orphan
                    normal_children.append(child)
                elif isinstance(child, Element):
                    text = False
                    child = child.orphan
                    normal_children.append(child)
                    process_element(child)
                else:
                    raise TypeError
                index+=1
            children = tuple(normal_children)

        for name in name_map:
            name_map[name]=tuple(name_map[name])

        self._children = children
        self._name_map = name_map
        self._free_prefix_map = free_prefix_map

    def to_libxml2(self, env=None, parent=None):
        if parent == None:
            raise NotImplementedError
        for child in self._children:
            child.to_libxml2(env, parent)
        return parent

    def is_orphan(self): return True

    def with_child(self, child, index):
        children = list(self._children)
        children[index]=child
        return Content(children)

    def __len__(self): return len(self._children)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return Content(self._children.__getitem__(key))
        else:
            return self._children[key]

    def __iter__(self): return self._children.__iter__()

    def __contains__(self, child):
        if isinstance(child, Element):
            self.content_map.__contains__(child)
        self._children.__contains__(child)

    def __eq__(self, other):
        return ((id(self) == id(other))
                or (isinstance(other, _OrphanedContent)
                    and (self._children == other._children)))

class _AdoptedContent(Content):
    """Implementation of Content with a parent."""

    __slots__ = ['_parent', '_orphan']
    parent = property(lambda self: self._parent)
    orphan = property(lambda self: self._orphan)
    string_value = property(lambda self: self._orphan.string_value)

    def __init__(self, orphan, parent):
        self._orphan = orphan
        self._parent = parent

    def is_orphan(self): return False

    def with_child(self, child, index):
        return self.parent.with_content(
            self.orphan.with_child(child, index)).content

    def __len__(self): return len(self._orphan)

    def __getitem__(self, key):
        if isinstance(key, slice):
            raise NotImplementedError
        if key < 0:
            index = key + len(self)
        else:
            index = key
        return self._orphan[key]._set_parent(self._parent, index)

    def __iter__(self):
        index = 0
        count = len(self)
        while index < count:
            yield self[index]
            index += 1

class Node(object):
    """Xml node abstract base class.

    Nodes are __eq__ and __ne__ comparable with other Nodes.

    Nodes should be __lt__, __le__, __gt__, and __ge__ comparable with
    other Nodes in the same document (according to document order as
    defined in XPath), but this is not currently implemented."""

    __slots__ = []

    name = property(lambda self: None,
                    doc="The expanded name or None.")
    ns = property(lambda self: None,
                  doc="The namespace uri or None.")
    qname = property(lambda self: None,
                     doc="The qualified name or None.")
    prefix = property(lambda self: None,
                      doc="The prefix part of the qualified name or None.")
    local_name = property(lambda self: None,
                          doc="The local part of the qualified name or None.")
    string_value = property(doc="The string value as defined in XPath.")
    parent = property(lambda self: None,
                      doc="The parent as defined in XPath or None.")
    orphan = property(doc="The node without a parent.")
    content = property(lambda self: None,
                         doc="The node's content (as a sequence) or None.")
    content_map = property(lambda self: None,
                           doc="A map of the node's element content (by name) or None.")

    def __top(self):
        result = self
        while result.parent != None: result = result.parent
        return result

    top = property(__top, doc="The oldest ancestor of this Node.")

    def __init__(self):
        raise NotImplementedError

    def is_orphan(self):
        raise NotImplementedError

    def replace(self, node):
        """Replace node with self in node's tree."""
        if not isinstance(node, _ChildMixin):
            raise NotImplementedError
        if not isinstance(self, (PI, Comment, Text, Element)):
            raise NotImplementedError
        return node.parent.content.with_child(self, node.index)[node.index]

    def to_libxml2(self):
        raise NotImplementedError

    def __str__(self):
        return self.serialize()

    def __hash__(self): raise NotImplemented
    def __eq__(self, other): raise NotImplemented
    def __ne__(self, other): return not (self == other)
    def __lt__(self, other): raise NotImplemented
    def __le__(self, other): raise NotImplemented
    def __gt__(self, other): raise NotImplemented
    def __ge__(self, other): raise NotImplemented

class _NamedMixin(object):
    # ISSUE: Python slots (currently?) have a "lay-out conflict"
    # restriction with multible inheritance, so this Mixin's slots
    # must be manually added to derived classes.
    #
    #__slots__ = ['_name', '_prefix']
    __slots__ = []
    name = property(lambda self: self._name)
    ns = property(lambda self: self._name.ns)
    prefix = property(lambda self: self._prefix)
    qname = property(lambda self: QName(self._prefix, self._name.local))
    local_name = property(lambda self: self._name.local)

    def __init__(self, prefix, name):
        self._prefix = prefix
        self._name = name

class _AdoptedNamedMixin(object):
    __slots__ = []
    name = property(lambda self: self.orphan.name)
    prefix = property(lambda self: self.orphan.prefix)
    qname = property(lambda self: self.orphan.qname)
    local_name = property(lambda self: self.orphan.local_name)

class _StringValueMixin(object):
    # ISSUE: Python slots currently have a "lay-out conflict"
    # restriction with multible inheritance, so this Mixin's slots
    # must be manually added to derived classes.
    #
    #__slots__ = ['_string_value']
    __slots__ = []
    string_value = property(lambda self: self._string_value)

    def __init__(self, string_value):
        self._string_value = string_value

class _AdoptedStringValueMixin(object):
    __slots__ = []
    string_value = property(lambda self: self.orphan.string_value)

class _ParentMixin(object):
    # ISSUE: Python slots currently have a "lay-out conflict"
    # restriction with multible inheritance, so this Mixin's slots
    # must be manually added to derived classes.
    #
    #__slots__ = ['_content']
    __slots__ = []

    def __init__(self, content):
        self._content = content

    content = property(
        lambda self: _AdoptedContent(self._content, self))

    content_map = property(lambda self: self.content.content_map)

class _AdoptedParentMixin(object):
    __slots__ = []

    content = property(
        lambda self: _AdoptedContent(self.orphan._content, self))

    content_map = property(lambda self: self.content.content_map)

class _OrphanedMixin(object):
    __slots__ = []
    parent = property(lambda self: None)
    orphan = property(lambda self: self)

    def is_orphan(self): return True

class _AdoptedMixin(object):
    __slots__ = ['_parent', '_orphan']
    parent = property(lambda self: self._parent)
    orphan = property(lambda self: self._orphan)

    def __init__(self, orphan, parent):
        self._orphan = orphan
        self._parent = parent

    def is_orphan(self): return False

    def serialize(self, encoding=None, indent=False):
        return self.orphan.serialize(encoding, indent)

class _ChildMixin(_AdoptedMixin):
    __slots__ = ['_index']
    index = property(lambda self: self._index)

    def __init__(self, orphan, parent, index):
        _AdoptedMixin.__init__(self, orphan, parent)
        self._index = index

    def to_libxml2(self):
        child = self.parent.to_libxml2().children
        index = 0
        while index < self._index:
            child = child.next
            index += 1
        return child

    def __eq__(self, other):
        return ((id(self) == id(other))
                or (isinstance(other, _ChildMixin)
                    and (self._index == other._index)
                    and (self._parent == other._parent)))

class Comment(Node):
    """An xml comment node."""

    __slots__ = []

    def __new__(cls, *rest, **krest):
        if (cls == Comment): cls = _OrphanedComment
        return Node.__new__(cls, *rest, **krest)

    def __init__(self, string_value):
        raise NotImplementedError

    def from_libxml2(comment, is_orphan):
        if is_orphan or (comment.parent == None):
            return Comment(comment.content)
        else:
            return from_libxml2(comment.parent, False).content[
                _libxml2_get_index(comment)]
    from_libxml2 = staticmethod(from_libxml2)

    def __hash__(self):
        return self.string_value.__hash__()

class _OrphanedComment(_StringValueMixin,
                       _OrphanedMixin, Comment):
    """Implementation of Comment without a parent."""

    __slots__ = ['_string_value'] # from _StringValueMixin

    def __init__(self, string_value):
        _StringValueMixin.__init__(self, _check_string_value(string_value))

    def _set_parent(self, parent, index):
        return _AdoptedComment(self, parent, index)

    def to_libxml2(self, env=None, parent=None):
        comment = libxml2.newComment(self.string_value)
        if parent != None: parent.addChild(comment)
        return comment

    def serialize(self, encoding=None, indent=False):
        n = None
        try:
            n = self.to_libxml2()
            return n.serialize(encoding, indent)
        finally:
            if n != None: n.freeNode()

    def __eq__(self, other):
        return ((id(self) == id(other))
                or (isinstance(other, _OrphanedComment)
                    and (self.string_value == other.string_value)))

class _AdoptedComment(_AdoptedStringValueMixin,
                      _ChildMixin, Comment):
    """Implementation of Comment with a parent."""

    __slots__ = []

    def __init__(self, orphan, parent, index):
        _ChildMixin.__init__(self, orphan, parent, index)

class PI(Node):
    """An xml processing instruction node."""

    __slots__ = []

    def __new__(cls, *rest, **krest):
        if (cls == PI): cls = _OrphanedPI
        return Node.__new__(cls, *rest, **krest)

    def __init__(self, name, string_value):
        raise NotImplementedError

    def from_libxml2(pi, is_orphan):
        if is_orphan or (pi.parent == None):
            return PI(pi.name, pi.content)
        else:
            return from_libxml2(pi.parent, False).content[
                _libxml2_get_index(pi)]
    from_libxml2 = staticmethod(from_libxml2)

class _OrphanedPI(_StringValueMixin, _NamedMixin,
                  _OrphanedMixin, PI):
    """Implementation of PI without a parent.

    ISSUE: http://www.w3.org/TR/REC-xml-names/ says "The default
    namespace can be set to the empty string. This has the same
    effect, within the scope of the declaration, of there being no
    default namespace". Is it also legal to have a prefix bound to no
    namespace?"""

    __slots__ = ['_string_value', # from _StringValueMixin
                 '_name', '_prefix'] # from _NamedMixin

    def __init__(self, name, string_value):
        prefix, name = _check_name(None, name)
        if name.ns != None: raise NameError
        _StringValueMixin.__init__(self, _check_string_value(string_value))
        _NamedMixin.__init__(self, None, name)

    def _set_parent(self, parent, index):
        return _AdoptedPI(self, parent, index)

    def to_libxml2(self, env=None, parent=None):
        pi = libxml2.newPI(self.local_name, self.string_value)
        if parent != None: parent.addChild(pi)
        return pi

    def serialize(self, encoding=None, indent=False):
        n = None
        try:
            n = self.to_libxml2()
            return n.serialize(encoding, indent)
        finally:
            if n != None: n.freeNode()

    def __eq__(self, other):
        return ((id(self) == id(other))
                or (isinstance(other, _OrphanedPI)
                    and (self._name == other._name)
                    and (self._prefix == other._prefix)
                    and (self._string_value == other._string_value)))

class _AdoptedPI(_AdoptedStringValueMixin, _AdoptedNamedMixin,
                 _ChildMixin, PI):
    """Implementation of PI with a parent."""

    __slots__ = []

    def __init__(self, orphan, parent, index):
        _ChildMixin.__init__(self, orphan, parent, index)

class Text(Node):
    """An xml text node."""

    __slots__ = []

    def __new__(cls, *rest, **krest):
        if (cls == Text): cls = _OrphanedText
        return Node.__new__(cls, *rest, **krest)

    def __init__(self, string_value):
        raise NotImplementedError

    def from_libxml2(text, is_orphan):
        if is_orphan or (text.parent == None):
            return Text(text.content)
        else:
            return from_libxml2(text.parent, False).content[
                _libxml2_get_index(text)]
    from_libxml2 = staticmethod(from_libxml2)

class _OrphanedText(_StringValueMixin,
                    _OrphanedMixin, Text):
    """Implementation of Text without a parent."""

    __slots__ = ['_string_value'] # from _StringValueMixin

    def __init__(self, string_value):
        string_value = _check_string_value(string_value)
        if len(string_value) == 0:
            raise EmptyTextError
        _StringValueMixin.__init__(self, string_value)

    def _set_parent(self, parent, index):
        return _AdoptedText(self, parent, index)

    def to_libxml2(self, env=None, parent=None):
        text = libxml2.newText(self.string_value)
        if parent != None: parent.addChild(text)
        return text

    def serialize(self, encoding=None, indent=False):
        n = None
        try:
            n = self.to_libxml2()
            return n.serialize(encoding, indent)
        finally:
            if n != None: n.freeNode()

    def __eq__(self, other):
        return ((id(self) == id(other))
                or (isinstance(other, _OrphanedText)
                    and (self._string_value == other._string_value)))

class _AdoptedText(_AdoptedStringValueMixin,
                   _ChildMixin, Text):
    """Implementation of Text with a parent."""

    __slots__ = []

    def __init__(self, orphan, parent, index):
        _ChildMixin.__init__(self, orphan, parent, index)

class NsDecl(Node):
    """An xml namespace declaration.

    In addition to acting as an xml node, an NsDecl acts as an ExpName
    factory."""

    __slots__ = []

    def __new__(cls, *rest, **krest):
        if (cls == NsDecl): cls = _OrphanedNsDecl
        return Node.__new__(cls, *rest, **krest)

    def __init__(self, prefix, ns):
        raise NotImplementedError

    def from_libxml2(nsd, is_orphan):
        if is_orphan:
            return NsDecl(nsd.name, nsd.content)
        else:
            # NOTE: libxml2 doesn't have a parent link for namespace declarations
            raise NotImplementedError
    from_libxml2 = staticmethod(from_libxml2)

    def __call__(self, local):
        """Create an ExpName with the give local part."""
        return ExpName(local, self._ns)

class _OrphanedNsDecl(_OrphanedMixin, NsDecl):
    """Implementation of NsDecl without a parent."""

    __slots__ = ['_prefix', '_ns']

    prefix = property(lambda self: self._prefix)
    ns = property(lambda self: self._ns)

    string_value = property(lambda self: self.ns or '')

    def __init__(self, prefix, ns):
        _check_name(prefix, ExpName('tmp', ns))
        self._prefix = prefix
        self._ns = ns

    def _set_parent(self, parent):
        return _AdoptedNsDecl(self, parent)

    def to_libxml2(self, env=None, parent=None):
        if parent == None:
            raise NotImplementedError
        return parent.newNs(self.ns, self.prefix)

    def serialize(self, encoding=None, indent=False):
        doc = None
        try:
            doc = libxml2.newDoc("1.0")
            element = doc.newChild(None, "tmp", None)
            nsd = element.newNs(self.ns, self.prefix)
            return nsd.serialize(encoding, indent)
        finally:
            if doc != None: doc.freeDoc()

    def __eq__(self, other):
        return ((id(self) == id(other))
                or (isinstance(other, _OrphanedNsDecl)
                    and (self._prefix == other._prefix)
                    and (self._ns == other._ns)))

class _AdoptedNsDecl(_AdoptedStringValueMixin, _AdoptedNamedMixin,
                     _AdoptedMixin, NsDecl):
    """Implementation of NsDecl with a parent."""

    __slots__ = []

    def __init__(self, orphan, parent):
        _AdoptedMixin.__init__(self, orphan, parent)

    def to_libxml2(self):
        raise NotImplementedError

    def __eq__(self, other):
        return (
            (id(self) == id(other))
            or (isinstance(other, _AdoptedNsDecl)
                and (self._orphan._prefix == other._orphan._prefix)
                and (self._parent == other._parent)))

class Attribute(Node):
    """An xml attribute node."""

    __slots__ = []

    def __new__(cls, *rest, **krest):
        if (cls == Attribute): cls = _OrphanedAttribute
        return Node.__new__(cls, *rest, **krest)

    def __init__(self, prefix, name, string_value):
        raise NotImplementedError

    def from_libxml2(attr, is_orphan):
        try:
            ns = attr.ns().content
            if ns == '': ns = None
            prefix = attr.ns().name
            name = ExpName(attr.name, ns)
        except libxml2.treeError:
            prefix = None
            name = ExpName(attr.name, None)
        if is_orphan or (attr.parent == None):
            return Attribute(prefix, name, attr.content)
        else:
            return from_libxml2(attr.parent, False).attrs[name]
    from_libxml2 = staticmethod(from_libxml2)

class _OrphanedAttribute(_StringValueMixin, _NamedMixin,
                         _OrphanedMixin, Attribute):
    """Implementation of Attribute without a parent."""

    __slots__ = ['_string_value', # from _StringValueMixin
                 '_name', '_prefix'] # from _NamedMixin

    def __init__(self, prefix, name, string_value):
        prefix, name = _check_name(prefix, name)
        if (name.ns != None) and (prefix == None):
            raise NameError
        _StringValueMixin.__init__(self, _check_string_value(string_value))
        _NamedMixin.__init__(self, prefix, name)

    def _set_parent(self, parent):
        return _AdoptedAttribute(self, parent)

    def to_libxml2(self, env=None, parent=None):
        if parent==None:
            raise NotImplementedError
        if self.prefix == None:
            return parent.newProp(self.local_name, self.string_value)
        else:
            return parent.newNsProp(_resolve_prefix(self.prefix, env),
                                    self.local_name,
                                    self.string_value)

    def serialize(self, encoding=None, indent=False):
        doc = None
        try:
            doc = libxml2.newDoc("1.0")
            element = doc.newChild(None, "tmp", None)
            if self.prefix != None:
                nsd = element.newNs(self.ns, self.prefix)
                attr = element.newNsProp(nsd,
                                         self.local_name,
                                         self.string_value)
            else:
                attr = element.newProp(self.local_name,
                                       self.string_value)
            return attr.serialize(encoding, indent)
        finally:
            if doc != None:
                doc.freeDoc()

    def __eq__(self, other):
        return ((id(self) == id(other))
                or (isinstance(other, _OrphanedAttribute)
                    and (self._name == other._name)
                    and (self._prefix == other._prefix)
                    and (self._string_value == other._string_value)))

class _AdoptedAttribute(_AdoptedStringValueMixin, _AdoptedNamedMixin,
                        _AdoptedMixin, Attribute):
    """Implementation of Attribute with a parent."""

    __slots__ = []

    def __init__(self, orphan, parent):
        _AdoptedMixin.__init__(self, orphan, parent)

    def to_libxml2(self):
        raise NotImplementedError

    def __eq__(self, other):
        return (
            (id(self) == id(other))
            or (isinstance(other, _AdoptedAttribute)
                and (self._orphan._name == other._orphan._name)
                and (self._parent == other._parent)))

class Element(Node):
    """An xml element node."""

    __slots__ = []

    attrs = property(
        doc="The element's attributes (as an AttrMap).")

    def __new__(cls, *rest, **krest):
        if (cls == Element): cls = _OrphanedElement
        return Node.__new__(cls, *rest, **krest)

    def auto(prefix, name, *fragments):
        """Helper constructor.

        This helper automatically sorts out the Comments, PIs, Text,
        Elements, Attributes, NsDecls, Content, strings, NsDeclMaps,
        AttrMaps, and sequences of the same among the given
        fragments. It ignores Nones.

        This helper also automatically inserts NsDecls in Elements to
        avoid prefix clashes, however it can't avoid prefix clashes
        among Attributes, nor deal with conflicting NsDecls."""

        content=[]
        attrs=[]
        nsds=[]
        def sort_fragment(fragment):
            if isinstance(fragment, (str, unicode)):
                content.append(Text(fragment))
            elif isinstance(fragment, Attribute):
                attrs.append(fragment)
            elif isinstance(fragment, NsDecl):
                nsds.append(fragment)
            elif isinstance(fragment, (Comment, PI, Text, Element)):
                content.append(fragment)
            elif isinstance(fragment, Content):
                content.extend(Content)
            elif isinstance(fragment, NsDeclMap):
                nsds.extend(fragment.values)
            elif isinstance(fragment, AttrMap):
                attrs.extend(fragment.values)
            elif fragment == None:
                pass
            else:
                sort_fragments(fragment)
        def sort_fragments(fragments):
            for fragment in fragments:
                sort_fragment(fragment)
        sort_fragments(fragments)

        # Try to resolve (some) prefix clashes:
        nsds = NsDeclMap(nsds)
        free_prefix_map = _FreePrefixMap(nsds)
        free_prefix_map.ref_prefix(prefix, name.ns)
        attrs = AttrMap(attrs)
        free_prefix_map.merge(attrs._free_prefix_map)
        for index, child in enumerate(content):
            try:
                if isinstance(child, Element):
                    free_prefix_map.merge(child._free_prefix_map)
            except PrefixClashError:
                content[index] = (
                    child.orphan.without_prefix_clash(free_prefix_map))
        return Element(prefix, name, content, attrs, nsds)
    auto = staticmethod(auto)

    def __init__(self, prefix, name,
                 content=(),
                 attributes=(),
                 ns_decls=()):
        raise NotImplementedError

    def __get_orphan_with_full_ns_env(self):
        def recurse(node):
            if ((node == None) or (isinstance(node, Root))):
                return {}
            else:
                env = recurse(node.parent)
                for prefix, nsd in node.orphan._ns_decls.items():
                    env[prefix] = nsd
                return env
        orphan = self.orphan
        ns_decls = recurse(self).values()
        if len(ns_decls) == len(orphan._ns_decls):
            return orphan
        else:
            return Element(orphan._prefix, orphan._name,
                           orphan._content, orphan._attrs,
                           NsDeclMap(ns_decls))

    orphan_with_full_ns_env = property(
        __get_orphan_with_full_ns_env,
        doc="The element without a parent, but with its ancestors' namespace declarations.")

    def __get_orphan_with_auto_ns_decls(self):
        orphan = self.orphan
        ns_decls = orphan._free_prefix_map.get_closed_ns_decls()
        if len(ns_decls) == len(orphan._ns_decls):
            return orphan
        else:
            return Element(orphan._prefix, orphan._name,
                           orphan._content, orphan._attrs, ns_decls)

    orphan_with_auto_ns_decls = property(
        __get_orphan_with_auto_ns_decls,
        doc="The element without a parent, and with all namespace prefixes bound.")

    def with_content(self, content):
        """Return a version of this node with different content"""
        raise NotImplementedError

    def from_libxml2(element, is_orphan):
        if is_orphan or (element.parent == None):
            children = []
            node = element.children
            while node:
                children.append(from_libxml2(node, True))
                node = node.next

            attributes = []
            node = element.get_properties()
            while node:
                attributes.append(Attribute.from_libxml2(node, True))
                node = node.next

            ns_decls = []
            try:
                node = element.nsDefs()
                while node:
                    ns_decls.append(NsDecl.from_libxml2(node, True))
                    node = node.next
            except libxml2.treeError:
                pass

            try:
                ns = element.ns().content
                if ns == '': ns = None
                prefix = element.ns().name
                name = ExpName(element.name, ns)
            except libxml2.treeError:
                prefix = None
                name = ExpName(element.name, None)

            return Element(prefix, name,
                           Content(children),
                           AttrMap(attributes),
                           NsDeclMap(ns_decls))
        else:
            return from_libxml2(element.parent, False).content[
                _libxml2_get_index(element)]
    from_libxml2 = staticmethod(from_libxml2)

    def relaxng_validate(self, ctxt):
        """Validate this element using the given validation context.

        The schemaContext should be a libxml2 validation context. Get
        this context from libxml2 directly, or use
        Root.to_xmlRelaxNGValidCtxt."""

        doc = None
        try:
            doc = Root((self.orphan_with_full_ns_env,)).to_libxml2()
            return doc.relaxNGValidateDoc(ctxt) == 0
        finally:
            if doc != None: doc.freeDoc()

    def c14n_serialize(self, with_comments=0, exclusive=0, prefixes=None):
        doc = None
        try:
            doc = Root((self.orphan_with_full_ns_env,)).to_libxml2()
            return doc.c14nMemory(None, exclusive, prefixes, with_comments)
        finally:
            if doc != None: doc.freeDoc()

class _OrphanedElement(_ParentMixin, _NamedMixin,
                       _OrphanedMixin, Element):
    __slots__ = ['_content', # from _ParentMixin
                 '_name', '_prefix', # from _NamedMixin
                 '_attrs', '_ns_decls',
                 '_free_prefix_map']

    ns_decls = property(lambda self: self._ns_decls)
    attrs = property(lambda self: self._attrs)
    string_value = property(lambda self: self._content.string_value)

    def __init__(self, prefix, name,
                 children=(),
                 attrs=(),
                 ns_decls=()):
        if isinstance(ns_decls, NsDeclMap):
            ns_decls = ns_decls.orphan
        else:
            ns_decls = NsDeclMap(ns_decls)
        free_prefix_map = _FreePrefixMap(ns_decls)

        prefix, name = _check_name(prefix, name)
        free_prefix_map.ref_prefix(prefix, name.ns)

        if isinstance(children, Content):
            content = children.orphan
        else:
            content = Content(children)
        free_prefix_map.merge(content._free_prefix_map)

        if isinstance(attrs, AttrMap):
            attrs = attrs.orphan
        else:
            attrs = AttrMap(attrs)
        free_prefix_map.merge(attrs._free_prefix_map)

        _NamedMixin.__init__(self, prefix, name)
        _ParentMixin.__init__(self, content)
        self._attrs = attrs
        self._ns_decls = ns_decls
        self._free_prefix_map = free_prefix_map

    def _set_parent(self, parent, index):
        return _AdoptedElement(self, parent, index)

    def with_content(self, content):
        return Element(self.prefix, self.name,
                       content, self.attrs, self.ns_decls)

    def without_prefix_clash(self, free_prefix_map):
        new_nsds = free_prefix_map.get_conflicting_ns_decls(self._free_prefix_map)
        if len(new_nsds) == 0:
            return self
        else:
            return _OrphanedElement(self.prefix,
                                    self.name,
                                    self.content,
                                    self.attrs,
                                    self.ns_decls.values() + new_nsds)

    def to_libxml2(self, env=None, parent=None, element=None):
        cleanup = False
        try:
            if element != None:
                pass
            elif parent != None:
                element = parent.newChild(None, self.local_name, None)
            else:
                element = libxml2.newNode(self.local_name)
                cleanup = True

            for nsd in self._ns_decls.itervalues():
                env = srfi1.alist_cons(
                    nsd.prefix,
                    nsd.to_libxml2(env, element),
                    env)
            element.setNs(_resolve_prefix(self.prefix, env))

            for attr in self._attrs.itervalues():
                attr.to_libxml2(env, element)

            self._content.to_libxml2(env, element)

            cleanup = False
            return element
        finally:
            if cleanup:
                # If we created the node, then we should free it if their
                # was a failure (like a MissingNsDeclError).
                element.freeNode()

    def libxml2_replace(self, orig):
        element = libxml2.newNode(self.local_name)
        orig.replaceNode(element)
        return self.to_libxml2(_libxml2_node_get_ns_env(orig.parent),
                               element.parent,
                               element)

    def serialize(self, encoding=None, indent=False):
        n = None
        try:
            try:
                n = self.to_libxml2()
                return n.serialize(encoding, indent)
            except MissingNsDeclError:
                ns_decls = self._free_prefix_map.get_closed_ns_decls()
                try:
                    prefix, ns_decl = ns_decls.iteritems().next()
                    ns = ns_decl.ns
                except StopIteration:
                    prefix, ns = None, None
                n = Element(prefix, ExpName("tmp", ns),
                            (self,), (), ns_decls).to_libxml2()
                return n.children.serialize(encoding, indent)
        finally:
            if n != None: n.freeNode()

    def __eq__(self, other):
        return ((id(self) == id(other))
                or (isinstance(other, _OrphanedElement)
                    and (self._name == other._name)
                    and (self._prefix == other._prefix)
                    and (self._content == other._content)
                    and (self._attrs == other._attrs)
                    and (self._ns_decls == other._ns_decls)))

class _AdoptedElement(_AdoptedParentMixin, _AdoptedNamedMixin,
                      _ChildMixin, Element):
    """Implementation of Element with a parent."""

    __slots__ = []

    ns_decls = property(lambda self: self._orphan.ns_decls._set_parent(self))
    attrs = property(lambda self: self._orphan.attrs._set_parent(self))
    string_value = property(lambda self: self._orphan.string_value)

    def __init__(self, orphan, parent, index):
        _ChildMixin.__init__(self, orphan, parent, index)

    def with_content(self, content):
        return self.parent.content.with_child(
            self.orphan.with_content(content),
            self.index)[self.index]

class Root(Node):
    """An xml root node."""

    # NOTE: There is only one representation of a root node, so this
    # abstract class isn't really necessary, but it makes the pydoc
    # documentation less confusing.

    __slots__ = []

    element = property(doc="The root element.")

    def __new__(cls, *rest, **krest):
        if (cls == Root): cls = _OrphanedRoot
        return Node.__new__(cls, *rest, **krest)

    def __init__(self, children=()):
        raise NotImplementedError

    def with_content(self, content):
        """Return a version of this node with different content"""
        raise NotImplementedError

    def from_libxml2(doc, is_orphan):
        children = []
        node = doc.children
        while node:
            children.append(from_libxml2(node, True))
            node = node.next
        return Root(Content(children))
    from_libxml2 = staticmethod(from_libxml2)

    def c14n_serialize(self, with_comments=0, exclusive=0, prefixes=None):
        doc = None
        try:
            doc = self.to_libxml2()
            return doc.c14nMemory(None, exclusive, prefixes, with_comments)
        finally:
            if doc != None: doc.freeDoc()

class _OrphanedRoot(_ParentMixin, _OrphanedMixin, Root):
    """An xml root node."""

    __slots__ = ['_content',  # from _ParentMixin
                 '_element_index']

    string_value = property(lambda self: self.element.string_value)
    element = property(
        lambda self: self.content[self._element_index])

    def __init__ (self, children=()):
        if isinstance(children, Content):
            # Make sure all descendents are orphans:
            content = children.orphan
        else:
            content = Content(children)

        element_index = None
        for index, child in enumerate(content):
            if isinstance(child, Text):
                raise TypeError
            if isinstance(child, Element):
                if element_index:
                    raise MultipleRootElementError
                else:
                    element_index = index
        if element_index == None:
            raise MissingRootElementError
        content._free_prefix_map.check_default()
        _ParentMixin.__init__(self, content)
        self._element_index = element_index

    def with_content(self, content):
        return Root(content)

    def to_libxml2(self):
        def convert_content(env, parent, content):
            for child in content:
                if isinstance(child, Element):
                    element = parent.newChild(None, child.local_name, None)
                    element_env = env
                    for nsd in child._ns_decls.itervalues():
                        element_env = srfi1.alist_cons(
                            nsd.prefix,
                            element.newNs(nsd.ns, nsd.prefix),
                            element_env)
                    element.setNs(_resolve_prefix(child.prefix, element_env))
                    for attr in child._attrs.itervalues():
                        if attr.prefix == None:
                            element.newProp(attr.local_name,
                                            attr.string_value)
                        else:
                            element.newNsProp(
                                _resolve_prefix(attr.prefix, element_env),
                                attr.local_name,
                                attr.string_value)
                    convert_content(element_env, element, child._content)
                else:
                    parent.addChild(child.to_libxml2())

        doc = libxml2.newDoc("1.0")
        self._content.to_libxml2(None, doc)
        return doc

    def serialize(self, encoding=None, indent=False):
        doc = None
        try:
            doc = self.to_libxml2()
            return doc.serialize(encoding, indent)
        finally:
            if doc != None: doc.freeDoc()

    def relaxng_validate(self, ctxt):
        """Validate this document using the given validation context.

        The schemaContext should be a libxml2 validation context. Get
        this context from libxml2 directly, or use
        Root.to_xmlRelaxNGValidCtxt."""
        doc = None
        try:
            doc = self.to_libxml2()
            return doc.relaxNGValidateDoc(ctxt) == 0
        finally:
            if doc != None: doc.freeDoc()

    def to_xmlRelaxNGValidCtxt(self):
        """DEPRECATED: use to_relaxng_ctxt"""
        return self.to_relaxng_ctxt()

    def to_relaxng_ctxt(self):
        """Create an libxml2 validation context using this Root as a schema.

        You can use the context in Root.relaxng_validate.
        """
        doc = None
        try:
            doc = self.to_libxml2()
            rngp = doc.relaxNGNewDocParserCtxt()
            rngs = rngp.relaxNGParse()
            return rngs.relaxNGNewValidCtxt()
        finally:
            if doc != None: doc.freeDoc()

    def __eq__(self, other):
        return ((id(self) == id(other))
                or (isinstance(other, _OrphanedRoot)
                    and (self._content == other._content)))

_from_libxml2_map = {
    'comment': Comment,
    'pi': PI,
    'text': Text,
    'namespace': NsDecl,
    'attribute': Attribute,
    'element': Element,
    'document_xml': Root}

def _libxml2_get_index(child):
    index = 0
    prev = child.prev
    while prev != None:
        index += 1
        prev = prev.prev
    return index

def _libxml2_ns_env_extend(env, element):
    nsd = element.nsDefs()
    while nsd:
        env = srfi1.alist_cons(nsd.name, nsd, env)
        libxml2_nsd = libxml2_nsd.next
    return env

def _libxml2_node_get_ns_env(node):
    if node == None:
        return None
    env = _get_libxml2_ns_env(node.parent)
    if node.type == 'element':
        return _libxml2_ns_env_extend(env, node)
    else:
        return env

def from_libxml2(node, is_orphan=False):
    try:
        return _from_libxml2_map[node.type].from_libxml2(node, is_orphan)
    except KeyError:
        raise Error('unrecognized libxml2 node type "%s".' % node.type)

def parse_file(filename):
    doc = None
    try:
        # ISSUE: libxml2 does not report the name of the file it was
        #   trying to open/parse.
        try:
            doc = libxml2.parseFile(filename)
            return from_libxml2(doc, True)
        except libxml2.parserError:
            raise ParseFileError(filename)
    finally:
        if doc != None: doc.freeDoc()

def parse_memory(buffer, size=None):
    doc = None
    try:
        doc = libxml2.parseMemory(buffer, size or len(buffer))
        return from_libxml2(doc, True)
    finally:
        if doc != None: doc.freeDoc()

def init_libxml2_debug():
    libxml2.debugMemory(1)

def quit_libxml2_debug():
    libxml2.relaxNGCleanupTypes()
    libxml2.cleanupParser()
    if libxml2.debugMemory(1) == 0:
        print "funkxml: libxml2 debugMemory: OK"
    else:
        print "funkxml: libxml2 debugMemory: Memory leak %d bytes" % (libxml2.debugMemory(1))
        libxml2.dumpMemory()

def examples():
    init_libxml2_debug()

    doc = """<?xml version="1.0"?>
<!--root comment-->
<everything xmlns="http://example.org/ns/foo" xmlns:foo="http://example.org/ns/foo" id="id attr"><!--element comment-->a little text<?in element pi?></everything>
<?in root pi?>
"""

    # To parse a document:
    root = parse_memory(doc)

    # Nodes print in standard xml syntax (this assertion assumes a
    # consistant serialization which is not guaranteed):
    assert doc == str(root)

    # Namespace:
    foo_ns = 'http://example.org/ns/foo'

    # Namespace declaration:
    foo = NsDecl(None, foo_ns)

    # A namespace declaration is an expanded name factory, for example:
    assert foo('a') == ExpName('a', foo_ns)

    # To synthesize a document:
    root = Root((Comment('root comment'),
                 Element(foo.prefix, foo('everything'),
                         (Comment('element comment'),
                          Text('a little text'),
                          PI('in', 'element pi')),
                         (Attribute(None, 'id', 'id attr'),),
                         (foo, NsDecl('foo', foo_ns))),
                 PI('in', 'root pi')))

    # The synthesized document is the same as the parsed one:
    assert root == parse_memory(doc)

    # To analyze the document, use the various accessors:
    assert root.parent == None
    assert root.content[0].parent == root
    assert root.element == root.content[1]
    assert root.content[0].string_value == 'root comment'
    assert isinstance(root.content[1], Element)

    # To extract a node from a document, get its orphan:
    assert root.content[0].orphan.parent == None

    # Nodes with different parents but the same content aren't the
    # same:
    assert root.content[0] != Comment('root comment')
    assert root.content[-1] != PI('in', 'root pi')
    # But as orphans, they are the same:
    assert root.content[0].orphan == Comment('root comment')
    assert root.content[-1].orphan == PI('in', 'root pi')

    # Wholesale parent:
    adoptees = root.content
    assert adoptees.parent == root
    assert not adoptees.is_orphan()

    # Wholesale orphans:
    orphans = root.content.orphan
    assert orphans[0].parent == None
    assert orphans[1].is_orphan()
    assert orphans[2].is_orphan()
    assert orphans.is_orphan()

    # Content is a python sequence:
    iter=root.content.__iter__()
    assert isinstance(iter.next(), Comment)
    assert isinstance(iter.next(), Element)
    assert isinstance(iter.next(), PI)
    for child in root.content:
        assert child.parent == root
    for child in root.content.orphan:
        assert child.parent == None

    # String values are as defined in XPath:
    assert root.string_value == 'a little text'
    assert root.content.string_value == 'a little text'
    assert root.content[0].string_value == 'root comment'
    assert root.content[1].string_value == 'a little text'

    # Element children work just like Root children:
    element = root.element
    children = element.content
    assert children.parent == element
    children = element.content.orphan
    assert children.parent == None
    assert children[0] == Comment('element comment')
    assert children[1] == Text('a little text')
    assert children[2] == PI('in', 'element pi')
    assert children[2] == children[-1]

    # Elements have a map of Attributes:
    assert element.attrs['id'].orphan == Attribute(None, 'id', 'id attr')
    assert element.attrs['id'].parent == element

    # Elements have a map of NsDecls:
    assert element.ns_decls[None].orphan == NsDecl(None, foo_ns)
    assert element.ns_decls[None].orphan == foo
    assert element.ns_decls[None].parent == element
    assert element.ns_decls['foo'].orphan == NsDecl('foo', foo_ns)
    assert element.ns_decls['foo'].parent == element

    # To share subtrees:
    cootie = Element(None, "cootie")
    element = Element(None, "bugs", (cootie, cootie, cootie))

    for shared_cootie in element.content:
        # According to XPath, documents don't share nodes, so in the
        # context of a parent, shared children aren't equal, ...
        assert cootie != shared_cootie
        # ... but outside of that context they are!
        assert cootie == shared_cootie.orphan

    # Children can also be accessed as a map:
    element = Element(None, 'a',
                      (Element(None, 'a', (Text('hello'),)),
                       Element(None, 'b', (Text('there'),)),
                       Element(None, 'a', (Text('world'),))))
    assert element.content_map.orphan['b'] == (Element(None, 'b', (Text('there'),)),)
    orphans = element.content_map.orphan
    assert orphans['a'] == (Element(None, 'a', (Text('hello'),)),
                            Element(None, 'a', (Text('world'),)))
    assert orphans['b'] == (Element(None, 'b', (Text('there'),)),)
    assert orphans.get('c', ()) == ()

    # Canonicalization:
    nsd0 = NsDecl('ns0', 'http://example.org/0')
    nsd1 = NsDecl('ns1', 'http://example.org/1')

    i0 = Element(None, 'inner', (Comment('inner 0'),), (), ())
    i1 = Element(None, 'inner', (Comment('inner 1'),), (), (nsd0,))
    i2 = Element(None, 'inner', (Comment('inner 2'),), (), (nsd0, nsd1))
    o0 = Element(None, 'outer', (i0,), (), (nsd0, nsd1))
    o1 = Element(None, 'outer', (i1,), (), (nsd0, nsd1))
    o2 = Element(None, 'outer', (i2,), (), (nsd0, nsd1))

    c14n_i = '<outer xmlns:ns0="http://example.org/0" xmlns:ns1="http://example.org/1"><inner></inner></outer>'
    c14n_e = '<outer><inner></inner></outer>'
    c14n_e_c0 = '<outer><inner><!--inner 0--></inner></outer>'
    c14n_e_c1 = '<outer><inner><!--inner 1--></inner></outer>'
    c14n_e_c2 = '<outer><inner><!--inner 2--></inner></outer>'

    # These three have different comments and namespace declarations,
    # so they serialize differently, ...
    assert o0 != o1
    assert o1 != o2
    assert o0.serialize() != o1.serialize()
    assert o1.serialize() != o2.serialize()

    # ... but they all canonicalize to the same document with c14n:
    assert o0.c14n_serialize() == c14n_i
    assert o1.c14n_serialize() == c14n_i
    assert o1.c14n_serialize() == c14n_i
    assert o0.c14n_serialize(exclusive=True) == c14n_e
    assert o1.c14n_serialize(exclusive=True) == c14n_e
    assert o2.c14n_serialize(exclusive=True) == c14n_e
    d0 = parse_memory(o0.c14n_serialize())
    d1 = parse_memory(o1.c14n_serialize())
    d2 = parse_memory(o2.c14n_serialize())
    assert d0 == d1
    assert d1 == d2
    # ... unless you use comments:
    assert o0.c14n_serialize(exclusive=True, with_comments=True) == c14n_e_c0
    assert o1.c14n_serialize(exclusive=True, with_comments=True) == c14n_e_c1
    assert o2.c14n_serialize(exclusive=True, with_comments=True) == c14n_e_c2

    # Test serialization with free prefix:
    assert str(Element("prefix", foo("name"))) == '<prefix:name/>'

    # Test full namepace environment:
    nsd0 = NsDecl(None, 'http://example.org/0')
    nsd1 = NsDecl('ns1', 'http://example.org/1')
    nsd2a = NsDecl('ns2', 'http://example.org/2a')
    nsd2b = NsDecl('ns2', 'http://example.org/2b')
    nsd3 = NsDecl('ns3', 'http://example.org/3')
    outer = Root ((Element(nsd1.prefix, nsd1('a'),
                           (Element(nsd1.prefix, nsd1('b'),
                                    (Element(nsd1.prefix, nsd1('c'),
                                             (),
                                             (),
                                             (nsd3, nsd2b)),),
                                    (),
                                    (nsd2a,)),),
                           (),
                           (nsd0, nsd1)),))
    inner = outer.element.content[0].content[0]

    assert inner.orphan != inner.orphan_with_full_ns_env
    assert (inner.orphan_with_full_ns_env
             == Element(nsd1.prefix, nsd1('c'),
                        (),
                        (),
                        (nsd0, nsd1, nsd2b, nsd3)))

    # RelaxNG Validation is a snap:

    schema = parse_memory("""<?xml version="1.0"?>
<element name="a" xmlns="http://relaxng.org/ns/structure/1.0">
  <text/>
</element>""")
    root = parse_memory('<?xml version="1.0"?>\n<a/>')

    assert root.relaxng_validate(schema.to_xmlRelaxNGValidCtxt())
    assert root.element.relaxng_validate(schema.to_xmlRelaxNGValidCtxt())

    #  Namespaces can be declared automatically:

    element = Element(None, ExpName('hello', 'http://example.org/ns'))
    manual_element = Element(None, ExpName('hello', 'http://example.org/ns'),
                             (), (), (NsDecl(None, 'http://example.org/ns'),))

    assert element.orphan_with_auto_ns_decls == manual_element

    # Helper methods make it easy to change content:

    doc = Root((Element(None, 'a', (Text('hello'),)),))
    element = doc.element
    new_element = element.with_content((Text('world'),))

    assert new_element.orphan == Element(None, 'a', (Text('world'),))
    assert new_element.parent == Root((Element(None, 'a', (Text('world'),)),))

    # Helper method makes it easier to build Elements:

    assert (Root((Comment('root comment'),
                  Element.auto(foo.prefix, foo('everything'),
                               Attribute(None, 'id', 'id attr'),
                               Comment('element comment'),
                               'a little text',
                               PI('in', 'element pi'),
                               foo,
                               NsDecl('foo', foo_ns)),
                 PI('in', 'root pi')))
            == Root((Comment('root comment'),
                     Element(foo.prefix, foo('everything'),
                             (Comment('element comment'),
                              Text('a little text'),
                              PI('in', 'element pi')),
                             (Attribute(None, 'id', 'id attr'),),
                             (foo, NsDecl('foo', foo_ns))),
                   PI('in', 'root pi'))))

    quit_libxml2_debug()

if __name__ == "__main__":
    examples()
