# Copyright (C) 2004  Anthony Carrico

"""
A module for performing OpenPGP operations.
"""

import os
import sys
from giftfile._osutil import PipeToChild, PipeFromChild, read_and_write_fds, \
    is_child
import re

# ISSUE: What is Mutt's "mutt_block_signals_system" all about?

class Error(Exception):
    def __init__(self, message=''):
        Exception.__init__(self, message)

_status_pattern = re.compile(
    r"""^\[GNUPG:\] (?P<keyword>\S+) ?(?P<rest>.*)$""", re.MULTILINE)

def _make_status_generator(text):
    "A generator to parse the gpg status text. Yields two values: the status keyword, and the rest of the status line. The rest string includes neither a space before the first status argument nor a trailing newline."

    for match in _status_pattern.finditer(text):
        yield match.group('keyword'), match.group('rest')

_SIG_CREATED_pattern = re.compile (r"""^(?P<type>\S+) (?P<pubkey_algo>\S+) (?P<hash_algo>\S+) (?P<class>\S+) (?P<timestamp>\S+) (?P<fingerprint>\S+)$""")

_KEY_CREATED_pattern = re.compile (r"""^(?P<type>\S+) (?P<fingerprint>\S+)$""")


# INV_RECP
#     INV_RECP <reason> <requested_recipient>
#         Issued for each unusable recipient. The reasons codes
#         currently in use are:
#           0 := "No specific reason given".
#           1 := "Not Found"
#           2 := "Ambigious specification"
#           3 := "Wrong key usage"
#           4 := "Key revoked"
#           5 := "Key expired"
#           6 := "No CRL known"
#           7 := "CRL too old"
#           8 := "Policy mismatch"
#           9 := "Not a secret key"
# 	 10 := "Key not trusted"

# NO_RECP
#   Issued when no recipients are usable.

_GOOD_or_BAD_SIG_pattern = re.compile(r"""^(?P<long_keyid>\S+) (?P<username>.+)$""")

_ERRSIG_pattern = re.compile(r"""^(?P<long_keyid>\S+) (?P<pubkey_algo>\S+) (?P<hash_algo>\S+) (?P<sig_class>\S+) (?P<timestamp>\S+) (?P<rc>.+)$""")

_VALIDSIG_pattern = re.compile(r"""^(?P<fingerprint>\S+) (?P<sig_creation_date>\S+) (?P<sig_timestamp>\S+) (?P<expire_timestamp>\S+) (?P<sig_version>\S+) (?P<reserved>\S+) (?P<pubkey_algo>\S+) (?P<hash_algo>\S+) (?P<sig_class>\S+) (?P<primary_key_fpr>\S+)$""")

class _colon_fields:
    """Enumeration of indexes in the gnupg colon listing format"""
    record_type = 0
    trust = 1
    key_length = 2
    algorithm = 3
    key_id = 4
    creation_date = 5
    expiration_date = 6
    serial_no = 7
    ownertrust = 8
    user_id = 9
    fingerprint = 9
    signature_class = 10
    key_capabilities = 11
    issuer_fingerprint = 12

######################################################################
## Verification

class SigBase(object):
    """Base class for signature info"""
    def __repr__(self):
        """
        Example output:
            <SigGood username="John Doe <john@example.net>", hash_algo='2',
                expired_key=False, ...>
        """
        def attribute_to_keyword((key, val)):
            return '%s=%s' % (key, repr(val))
        name = type(self).__name__
        keywords = map(attribute_to_keyword, self.__dict__.iteritems())
        return '<%s %s>' % (name, ', '.join(keywords))

class SigSpec(SigBase):
    """OpenPGP signature parameters."""
    def __init__(self, pubkey_algo, hash_algo, sig_class):
        self.pubkey_algo = pubkey_algo
        "currently a string. Ex: '17'"
        self.hash_algo = hash_algo
        "currently a string. Ex: '2'"
        self.sig_class = sig_class
        "currently a string. Ex: '00'"

class SigCommon(SigSpec):
    """Common details about a signature"""
    def __init__(self, pubkey_algo, hash_algo, sig_class, timestamp, fingerprint):
        SigSpec.__init__(self, pubkey_algo, hash_algo, sig_class)
        self.timestamp = timestamp
        """Signature creation time in seconds after the epoch.

        Currently a string. Ex: '1075490244'"""
        self.fingerprint = fingerprint
        "Currently a string. Ex: 'BCA6936F8278FAE25E24EBFE50D781FB74937222'"

class SigNew(SigCommon):
    """Details about a created signature"""

    def __init__(self, pubkey_algo, hash_algo, sig_class, timestamp, fingerprint, sig_type):
        SigCommon.__init__(self, pubkey_algo, hash_algo, sig_class, timestamp, fingerprint)
                # Detached, cleartext, or standard type
        # ISSUE: checkup on gpg usage: "only the first character
        # should be checked" -- does that mean gpg might return more
        # than one in this field?
        self.sig_type = sig_type
        "Detached, cleartext, or standard. Currently a string 'D', 'C', or 'S'."

class SigGood(SigCommon):
    """Details about a signature verified to be ok.

    The signature with the keyid is good, however it might be expired,
    or the key might be expired or revoked, or you might not trust the
    key."""

    def __init__(self, keyid, username, expired, expired_key, revoked_key):
        # NOTE: SigGood is initialized in several steps, so
        # __init__ doesn't call SigCommon.__init__
        self.keyid = keyid
        "Currently a string. Ex. '50D781FB74937222'."
        self.username = username # string
        "String. Ex. 'Jane Doe <jane-doe@bogus>'"
        self.expired = expired
        "Signature expired? boolean."
        self.expired_key = expired_key
        "Key expired? boolean"
        self.revoked_key = revoked_key
        "Key revoked? boolean"
        # ISSUE: Notations and policy_url aren't
        self.notations = {}
        "Table of notation name (string) to data (string). CURRENTLY NOT % DECODED!"
        self.policy_url = None
        "String or None. CURRENTLY NOT % DECODED!"
        self.trust = 'TRUST_UNDEFINED'
        """Currently a string.

        One of 'TRUST_UNDEFINED' 'TRUST_NEVER' 'TRUST_MARGINAL'
        'TRUST_FULLY', or 'TRUST_ULTIMATE'."""
        self.expire_timestamp = None
        """Signature expiration time in seconds after the epoch.

        Currently a string or None"""
        self.sig_version = None
        "Currently a string. Ex: '3'"
        self.primary_key_fpr = None
        "Currently a string."

    def _init2 (self, expire_timestamp, sig_version, primary_key_fpr):
        # Signature expiration time in seconds after the epoch:
        self.expire_timestamp = expire_timestamp # string or None
        self.sig_version = sig_version # string. Ex: '3'
        self.primary_key_fpr = primary_key_fpr # string.

    def _init3 (self, trust):
        # Trust values: 'TRUST_UNDEFINED' 'TRUST_NEVER'
        # 'TRUST_MARGINAL' 'TRUST_FULLY' 'TRUST_ULTIMATE'
        self.trust = trust # string

class SigBad(SigBase):
    """Details about a signature NOT verified to be ok."""

    def __init__(self, keyid, username):
        self.keyid = keyid
        "Currently a string. Ex. '50D781FB74937222'."
        self.username = username
        "String. Ex. 'Jane Doe <jane-doe@bogus>'"

class SigError(SigSpec):
    """Details about a signature that could not be verified."""

    def __init__(self, keyid, pubkey_algo, hash_algo, sig_class, timestamp):
        SigSpec.__init__(self, pubkey_algo, hash_algo, sig_class)
        self.keyid = keyid
        "Currently a string. Ex. '50D781FB74937222'."
        self.timestamp = timestamp
        "Signature creation time in seconds since epoch.  Ex: '1075490244'"

class SigUnknownAlgorithm(SigError):
    """Details about a signature with an unknown algorithm."""

    def __init__(self, keyid, pubkey_algo, hash_algo, sig_class, timestamp):
        SigError.__init__(self, keyid, pubkey_algo, hash_algo, sig_class, timestamp)

class SigMissingKey(SigError):
    """Details about a signature with a missing public key."""

    def __init__(self, keyid, pubkey_algo, hash_algo, sig_class, timestamp):
        SigError.__init__(self, keyid, pubkey_algo, hash_algo, sig_class, timestamp)

class Context(object):
    """Some context for using OpenPGP."""

    BASE_GPG_OPTIONS = [
        '--no-auto-check-trustdb',
        '--quiet',
        '--no-permission-warning',
    ]

    def __exec_gpg (self, args):
        e = os.execlp
        if self.gpg_pathname:
            e = os.execl
        op = self.gpg_pathname or 'gpg'
        auto_args = self.BASE_GPG_OPTIONS + [
            '--logger-fd', str(self.log_file.fileno()),
        ]
        if self.homedir:
            auto_args += ['--homedir', self.homedir]
        if self.keyring:
            auto_args += ['--keyring', self.keyring, '--no-default-keyring']
        if self.keyserver:
            auto_args += ['--keyserver', self.keyserver]
        if self.auto_key_retrieve:
            auto_args += ['--keyserver-options', 'auto-key-retrieve']
        # NOTE: auto args are prepended since user args may contain "--"
        e (op, op, *(auto_args + args))

    def __init__ (self, gpg_pathname=None, homedir=None, keyring=None,
            keyserver=None, auto_key_retrieve=False, log_file=sys.stderr):
        """
        TODO: add support for --secret-keyring
        """
        self.gpg_pathname = gpg_pathname
        "The full pathname (string) of the gpg program, or None. If None, search for gpg in the path."
        self.homedir = homedir
        "The gpg home directory or None (usually defaults to ~/.gnupg)."
        self.keyring = keyring
        "If set, given keyring will be used exclusively for verification"
        self.keyserver = keyserver
        "Optional keyserver."
        self.auto_key_retrieve = auto_key_retrieve
        "Fetch keys from keyserver automatically on verify."
        self.log_file = log_file
        "Python file object for log output."

    def gen_key(self, params_list):
        """Generate a new key pair.

        ISSUE: shouldn't this interface be abstracted from GnuPG?

        Each item in params_list is a dictionary of parameters (see "Unattended
        key generation" in DETAILS from the gnupg
        distribution). Currently control statements and subkeys are
        not supported.

        The result is a list of corresponding fingerprints.

        ISSUE: Ideally the result list would contain None or some
        error if a parameter set fails, but currently it is not
        possible to tell which key fails, so the result list will be
        short in that case."""

        head_key = "Key-Type"

        def make_param(key, value):
            return key + ': ' + value

        def make_tail_param(key, value):
            if key == head_key:
                return '#'
            else:
                return make_param(key, value)

        batch_data = ('\n'.join(map(lambda params: '\n'.join(
            [make_param(head_key, params[head_key])]
            + map(lambda p: make_tail_param(p[0], p[1]), params.items())),
                                  params_list))
                      + '\n')

        in_pipe = PipeToChild()
        out_pipe = PipeFromChild()
        status_pipe = PipeFromChild()

        pid = os.fork()
        if is_child(pid):
            in_fd = in_pipe.extract_child_fd()
            os.dup2(in_fd, 0)
            os.close(in_fd)
            self.__exec_gpg(['--enable-special-filenames',
                             '--status-fd', repr(status_pipe.extract_child_fd()),
                             '--gen-key',
                             '--batch',
                             '-o', '-&' + repr(out_pipe.extract_child_fd()),
                             '-'])
        else:
            status_fd = status_pipe.extract_parent_fd()
            out_fd = out_pipe.extract_parent_fd()
            read_table = read_and_write_fds(
                [status_fd, out_fd],
                {in_pipe.extract_parent_fd() : batch_data})
            ret_pid, p_status = os.waitpid(pid, 0)
            assert (pid == ret_pid)

            # Note: There doesn't seem to be any output, but keep the
            # out_fd for future debugging.
            #
            # out_text = read_table[out_fd]

            status_text = read_table[status_fd]

            results = []

            for keyword, rest in _make_status_generator(status_text):
                if keyword == "KEY_CREATED":
                    m = _KEY_CREATED_pattern.match(rest)
                    assert (m.group('type') == 'P')
                    results.append(m.group('fingerprint'))

            return results

    def send_keys(self, names):
        """Send keys to keyserver

        names: list of key names (key-ID, fingerprint, user-ID, etc.)

        ISSUE: gpg does not provide any status for this operation.  No
            error is generated if a name does not exist in the keyring.
        """

        pid = os.fork()
        if is_child(pid):
            self.__exec_gpg(['--send-keys'] + names)
        else:
            ret_pid, p_status = os.waitpid(pid, 0)
            assert (pid == ret_pid)

    def list_keys(self, name=None, secret_keys=False):
        """List keys in keyring

        name: optional search name (key-ID, fingerprint, user-ID, etc.)
        secret_keys: if True, list secret keys, otherwise list public keys

        Returns a list of fingerprints.
        """
        out_pipe = PipeFromChild()
        pid = os.fork()
        if is_child(pid):
            # ISSUE: --list-keys does not respect --output option
            #   See http://bugs.gnupg.org/cgi-bin/gnatsweb.pl?cmd=view&pr=300.
            #   Using dup causes our prints to get scooped up.
            out_fd = out_pipe.extract_child_fd()
            os.dup2(out_fd, 1)
            os.close(out_fd)
            args = [
                '--with-colons',
                '--with-fingerprint',
                '--fast-list-mode',
            ]
            if secret_keys:
                args += ['--list-secret-keys']
            else:
                args += ['--list-keys']
            if name:
                args.append(name)
            self.__exec_gpg(args)
        else:
            out_fd = out_pipe.extract_parent_fd()
            read_table = read_and_write_fds([out_fd], {})
            ret_pid, p_status = os.waitpid(pid, 0)
            assert (pid == ret_pid)
            out_text = read_table[out_fd]
            lines = out_text.splitlines()
            result = []
            for line in lines:
                fields = line.split(':')
                if fields[_colon_fields.record_type] == 'fpr':
                    result.append(fields[_colon_fields.fingerprint])
            return result

    def detach_sign(self, data, signers=None):
        """Make a detached signature.

        Gnupg gets control of standard in/out/error for user dialog.

        Signers is a sequence of ids to sign with (the format is not
        yet specified), None will sign with the default id.

        Returns a list of SigNew objects, and the signature or None."""

        # Create some pipes to connect to gnupg:
        data_pipe = PipeToChild()
        sig_pipe = PipeFromChild()
        status_pipe = PipeFromChild()

        # Fork:
        pid = os.fork()
        if is_child(pid):
            args = []
            # ISSUE: What is --local-user vs. --default-key?
            if signers:
                for s in signers:
                    args = args + ['--local-user', s]
            args = args + ['--enable-special-filenames',
                           '--status-fd', repr(status_pipe.extract_child_fd()),
                           '--detach-sign',
                           '-o', "-&" + repr(sig_pipe.extract_child_fd()),
                           "--",
                           "-&" + repr(data_pipe.extract_child_fd())]
            self.__exec_gpg(args)
        else:
            status_fd = status_pipe.extract_parent_fd()
            sig_fd = sig_pipe.extract_parent_fd()
            read_table = read_and_write_fds ([status_fd, sig_fd],
                                             {data_pipe.extract_parent_fd():data})

            ret_pid, p_status = os.waitpid(pid, 0)
            assert (pid == ret_pid)

            status_text = read_table[status_fd]

            #print
            #print "status: " + status_text

            sns = []

            for keyword, rest in _make_status_generator(status_text):
                if keyword == "SIG_CREATED":
                    m = _SIG_CREATED_pattern.match(rest)
                    sns.append(SigNew (m.group('pubkey_algo'),
                                 m.group('hash_algo'),
                                 m.group('class'),
                                 m.group('timestamp'),
                                 m.group('fingerprint'),
                                 m.group('type')))

            if len(sns) > 0:
                sig = read_table[sig_fd]
                return sns, sig
            else:
                return sns, None

    def detach_verify(self, sig, data):
        """Verify that the data was signed by the sig.

        Return a list of SigGood, SigBad, and/or SigError elements,
        one for each signature in sig."""

        # Create some pipes to connect to gnupg:
        data_pipe = PipeToChild()
        sig_pipe = PipeToChild()
        status_pipe = PipeFromChild()

        # Fork:
        pid = os.fork()
        if is_child(pid):
            self.__exec_gpg(
                ['--enable-special-filenames',
                 '--status-fd', repr(status_pipe.extract_child_fd()),
                 '--verify',
                 "--",
                 "-&" + repr(sig_pipe.extract_child_fd()),
                 "-&" + repr(data_pipe.extract_child_fd())])
        else:
            status_fd = status_pipe.extract_parent_fd()
            read_table = read_and_write_fds (
                [status_fd],
                {sig_pipe.extract_parent_fd():sig,
                 data_pipe.extract_parent_fd():data,})

            ret_pid, p_status = os.waitpid(pid, 0)
            assert (pid == ret_pid)

            status_text = read_table[status_fd]

            # print
            # print "status: " + status_text

            verifications = []
            cv = None # current verification
            cn = None # current notation

            for keyword, rest in _make_status_generator(status_text):
                if (keyword == "GOODSIG"):
                    m = _GOOD_or_BAD_SIG_pattern.match(rest)
                    if not m:
                        raise Error()
                    else:
                        cv = SigGood(m.group('long_keyid'),
                                     m.group('username'),
                                     False,
                                     False,
                                     False)
                        verifications.append(cv)
                        cn = None

                elif (keyword == "EXPSIG"):
                    m = _GOOD_or_BAD_SIG_pattern.match(rest)
                    if not m:
                        raise Error()
                    else:
                        cv = SigGood(m.group('long_keyid'),
                                     m.group('username'),
                                     True,
                                     False,
                                     False)
                        verifications.append(cv)
                        cn = None

                elif (keyword == "EXPKEYSIG"):
                    m = _GOOD_or_BAD_SIG_pattern.match(rest)
                    if not m:
                        raise Error()
                    else:
                        cv = SigGood(m.group('long_keyid'),
                                     m.group('username'),
                                     False,
                                     True,
                                     False,)
                        verifications.append(cv)
                        cn = None

                elif (keyword == "REVKEYSIG"):
                    m = _GOOD_or_BAD_SIG_pattern.match(rest)
                    if not m:
                        raise Error()
                    else:
                        cv = SigGood(m.group('long_keyid'),
                                     m.group('username'),
                                     False,
                                     False,
                                     True)
                        verifications.append(cv)
                        cn = None

                elif (keyword == "BADSIG"):
                    m = _GOOD_or_BAD_SIG_pattern.match(rest)
                    if not m:
                        raise Error()
                    else:
                        cv = SigBad(m.group('long_keyid'),
                                    m.group('username'))
                        verifications.append(cv)
                        cn = None

                elif (keyword == "ERRSIG"):
                    m = _ERRSIG_pattern.match(rest)
                    if not m:
                        raise Error()
                    else:
                        # TODO: enumerate these magic numbers
                        if (m.group('rc') == '4'):
                            cv = SigUnknownAlgorithm(m.group('long_keyid'),
                                                     m.group('pubkey_algo'),
                                                     m.group('hash_algo'),
                                                     m.group('sig_class'),
                                                     m.group('timestamp'))
                        elif (m.group('rc') == '9'):
                            cv = SigMissingKey(m.group('long_keyid'),
                                               m.group('pubkey_algo'),
                                               m.group('hash_algo'),
                                               m.group('sig_class'),
                                               m.group('timestamp'))
                        else:
                            cv = SigError(m.group('long_keyid'),
                                          m.group('pubkey_algo'),
                                          m.group('hash_algo'),
                                          m.group('sig_class'),
                                          m.group('timestamp'))

                        verifications.append(cv)
                        cn = None

                # ISSUE: It looks like gpg may issue the NO_PUBKEY keyword
                # following ERRSIG, but I'm not sure in what case, or what
                # to do with it, so I'll ignore it for now.

                elif (keyword == "NOTATION_NAME"):
                    # ISSUE: not doing anything to deal with %XX escapes in notation name.
                    if not (cv and isinstance(cv, SigGood)):
                        raise Error()
                    else:
                        cn = rest
                        cv.notations [cn] = ''

                elif (keyword == "NOTATION_DATA"):
                    # ISSUE: not doing anything to deal with %XX escapes in notation data.
                    if not (cv and cn):
                        raise Error()
                    else:
                        cv.notations [cn] += ''

                elif (keyword == "POLICY_URL"):
                    # ISSUE: not doing anything to deal with %XX escapes in the policy url.
                    if not (cv and cn):
                        raise Error()
                    else:
                        cv.policy_url = ''

                elif (keyword == "VALIDSIG"):
                    m = _VALIDSIG_pattern.match(rest)
                    if not (m and cv and isinstance(cv, SigGood)):
                        raise Error()
                    else:
                        # Note 'sig_creation_date' is redundant with
                        # sig_timestamp, so we ignore it.
                        SigCommon.__init__(cv,
                                           m.group('pubkey_algo'),
                                           m.group('hash_algo'),
                                           m.group('sig_class'),
                                           m.group('sig_timestamp'),
                                           m.group('fingerprint'))

                        expire_timestamp = m.group('expire_timestamp')
                        if (expire_timestamp == '0'):
                            expire_timestamp = None
                        cv._init2(expire_timestamp,
                                  m.group('sig_version'),
                                  m.group('primary_key_fpr'))

                elif (keyword == "KEYREVOKED"):
                    if not (m and cv and isinstance(cv, SigGood)):
                        raise Error()
                    else:
                        cv.revoked_key = True

                elif (keyword == "KEYEXPIRED"):
                    if not (m and cv and isinstance(cv, SigGood)):
                        raise Error()
                    else:
                        cv.expired_key = True

                elif ((keyword == "TRUST_UNDEFINED")
                      or (keyword == "TRUST_NEVER")
                      or (keyword == "TRUST_MARGINAL")
                      or (keyword == "TRUST_FULLY")
                      or (keyword == "TRUST_ULTIMATE")):
                    if not cv:
                        raise Error()
                    else:
                        cv.trust = keyword

                elif (keyword == "NODATA"):
                    # ISSUE: meaning of NODATA is ambiguous
                    #   This can occur when using --auto-key-retrieve and
                    #   the key lookup fails, which can't be distinguished
                    #   from NODATA error during message processing.  See
                    #   <http://bugs.gnupg.org/cgi-bin/gnatsweb.pl?cmd=view&pr=297>.
                    #   If auto key retrieve is enabled, we'll assume this
                    #   NODATA was due to a failed key lookup, and ignore it.
                    #   Note that SigMissingKey will be reported.
                    if not self.auto_key_retrieve:
                        raise Error

                elif (keyword == "UNEXPECTED"):
                    # ISSUE: Could there be some good signatures processed
                    # before or after "UNEXPECTED"?
                    raise Error()

            return verifications


def _test():
    from os.path import expanduser, exists
    gen_key_test_homedir = expanduser('~/tmp/gen_key_test')
    if exists(gen_key_test_homedir):
        ctxt = Context(homedir = gen_key_test_homedir)
        fprs = ctxt.gen_key([{"Key-Type":"DSA",
                             "Key-Length":"1024",
                             "Name-Real":"John Doe",
                             "Name-Email":"jdoe@example.org"},
                            {"Key-Type":"Bogus",
                             "Key-Length":"1024",
                             "Name-Real":"John Doe",
                             "Name-Email":"jdoe@example.org"}])
        print "fingerprints:"
        print fprs

if __name__ == '__main__':
    _test()
