# Copyright (C) 2004  Anthony Carrico

"""Python implementation of some of  Olin Shivers' linked list library.

Reference: http://srfi.schemers.org/srfi-1/srfi-1.html

Uses two-tuples as cons cells."""

def find(pred, l):
    """Return the first element of l that satisfies predicate pred; False if no element does.

    find(even, (3, (1, (4, (1, (5, (9, None))))))) => 4

    Note that find has an ambiguity in its lookup semantics -- if find
    returns False, you cannot tell (in general) if it found a False
    element that satisfied pred, or if it did not find any element at
    all. In many situations, this ambiguity cannot arise -- either the
    srfi1-list being searched is known not to contain any False
    elements, or the srfi1-list is guaranteed to have an element
    satisfying pred. However, in cases where this ambiguity can arise,
    you should use find_tail instead of find -- find_tail has no such
    ambiguity."""

    while l:
        if pred(l[0]): return l[0]
        else: l = l[1]
    return False

def find_tail(pred, l):
    """Return the first pair of l whose car satisfies pred. If no pair does, return False.

    find_tail can be viewed as a general-predicate variant of the member function.

    find_tail(even, (3, (1, (37, (-8, (-5, (0, (0, None)))))))) => (-8, (-5, (0, (0, None))))
    find_tail(even, (3, (1, (37, (-5, None))))) => #f

    Find-tail is essentially drop-while, where the sense of the
    predicate is inverted: Find-tail searches until it finds an
    element satisfying the predicate; drop-while searches until it
    finds an element that doesn't satisfy the predicate."""
    while l:
        if pred(l[0]): return l
        else: l = l[1]
    return False

def assoc(key, alist, eq_pred=None):
    """Find the first pair in alist whose car field is key.

    Return a pair or False.

    alist must be an association list -- a srfi1-list of pairs. find
    the first pair in alist whose car field is key, and return that
    pair. If no pair in alist has key as its car, then False is
    returned. assoc uses =, or the given equality predicate.  to
    compare key with the car fields of the pairs in alist."""

    if eq_pred:
        while alist:
            if eq_pred(key, alist[0][0]):
                return alist[0]
            else:
                alist = alist[1]
    else:
        while alist:
            if key == alist[0][0]:
                return alist[0]
            else:
                alist = alist[1]
    return False

def alist_cons(key, datum, alist):
    """Cons a new alist entry mapping key -> datum onto alist."""

    return ((key, datum), alist)



