# Copyright (C) 2004  John Belmonte

"""Status codes for the giftpool protocol

The giftpool protocol uses codes to indicate the status or result of
requests, queries, etc.

terminology:

    status code - machine-readable status identifier
    status reason - human-readable description of status code
    status message - combination of status code and reason

status message synopsis:

    1.2 Payment Not Received

Status codes are represented as a pair of integers having up to three digits
and no leading zeros, separated by a period.  This representation was
influenced by RFC 3463.

Status codes are devided into classes so that, should a client not
recognize a specific code, it can still operate according to the general
meaning of the class.  The defined classes are:

    1.XXX   Processing
    2.XXX   Successful
    4.XXX   Failed


ISSUE: module should be renamed to giftpool_status

ISSUE: how specific should status codes be?
    Can we rely on context?  For example, should there be a PAYMENT_BOUNCED
    and REMITTANCE_BOUNCED or just a BOUNCED?

NOTE: once released, existing codes can never be changed.  Only additions
    may be made.
"""

# Enumeration of status codes
PROCESSING                  = '1.0'
NEW_USER                    = '1.1'
PAYMENT_NOT_RECEIVED        = '1.2'
PAYMENT_NOT_CLEARED         = '1.3'
REMITTANCE_ISSUED           = '1.4'
OK                          = '2.0'
FAILED                      = '4.0'
PAYMENT_BOUNCED             = '4.1'
INSUFFICIENT_FUNDS          = '4.2'
REMITTANCE_EXPIRED          = '4.3'
REMITTANCE_DELIVERY_FAILED  = '4.4'
INVALID_RESOURCE            = '4.5'
TRANSACTION_TOO_SMALL       = '4.6'

# Mapping of status code to reason
reasons = {
    PROCESSING                  : 'Processing',
    NEW_USER                    : 'New User',
    PAYMENT_NOT_RECEIVED        : 'Payment Not Received',
    PAYMENT_NOT_CLEARED         : 'Payment Not Cleared',
    REMITTANCE_ISSUED           : 'Remittance Issued',
    OK                          : 'OK',
    FAILED                      : 'Failed',
    PAYMENT_BOUNCED             : 'Payment Bounced',
    INSUFFICIENT_FUNDS          : 'Insufficient Funds',
    REMITTANCE_EXPIRED          : 'Remittance Expired',
    REMITTANCE_DELIVERY_FAILED  : 'Remittance Delivery Failed',
    INVALID_RESOURCE            : 'Invalid Resource',
    TRANSACTION_TOO_SMALL       : 'Transaction Too Small',
}

def get_message(code):
    """Return status message given a status code

    Synopsis:
        >>> transaction_status.get_message(transaction_status.PROCESSING)
        '1.0 Processing'
    """
    return '%s %s' % (code, reasons[code])

def get_code(message):
    """Return status code given a status message

    Synopsis:
        >>> transaction_status.get_id('1.0 Processing')
        '1.0'
    """
    return message.split(' ', 1)[0]

# TODO: remove references to this deprecated name
get_id = get_code

def get_class_code(code_or_message):
    """Return status code of class given a status code or message

    Synopsis:
        >>> transaction_status.get_class_code('1.2')
        '1.0'
        >>> transaction_status.get_class_code('1.2 Payment Not Received')
        '1.0'
    """
    code = get_code(code_or_message)
    return code.split('.', 1)[0] + '.0'
