# Copyright (C) 2004  John Belmonte

"""Wrapper for libxml2 objects providing an xmltramp-like interface

Deviations from the xmltramp interface:
    * Namespace interface changes:
        * does not implement __getattr__ (provided in optional derived class)
    * Element interface changes:
        * initializer signature changed (synthesis of XML not supported)
        * mutation not supported
        * does not implement __getattr__
        * does not implement __hasattr__, use has_element method
        * __repr__ is verbose and suitable for eval.  No extensions provided.
            See repr module if you want a shorter version.
        * __str__ serializes the element.  Use .content attribute for content.
        * ._name and ._dir attributes renamed to .name and .elements
        * .name always returns a (ns, name) tuple, where ns may be None
        * added serialize() method
        * xmltramp binds unqualified names to the default namespace of
            the document, which is a design flaw because the default
            namespace is a rendering detail of the document.  Instead,
            this interface binds unqualified names to no namespace, and
            allows the user to override this with a specific namespace.

Variatons to try:
    * unify attribute and element access - name generators can be in attribute
        space, element space, or both;  remove __call__ access;
        add __contains__ access
"""

from types import StringTypes
import libxml2

def _qualified_class_name(obj):
    module, name = obj.__class__.__module__, obj.__class__.__name__
    prefix = ''
    if module != '__main__':
        prefix = module + '.'
    return prefix + name

def _ns(node):
    try:
        return node.ns().content
    except libxml2.treeError:
        pass

def _normalizeName(name, default_namespace=None):
    if isinstance(name, tuple):
        return name
    else:
        return (default_namespace, name)


class Namespace(object):
    """XML namespace which can generate expanded element and attribute names"""
    __slots__ = ['_uri']
    def __init__(self, uri): self._uri = uri
    def __getitem__(self, name): return(self._uri, name)


class Element(object):
    """XML element

    NOTE: can derive from this, adding caching, etc.
    """

    __slots__ = ['_doc', '_elm', '_owner', 'default_namespace']

    def __init__(self, xml, default_namespace=None, owner=False):
        """
        xml: XML text or libxml2.xmlElement/Doc instance.  In latter case
            references to the input object are kept.  This implies that input
            object mutations will be reflected by this object, and that it's
            not ok to use this object after the input object is freed.
        default_namespace: the namespace to use for unqualified names.  It
            may be read and changed after init via attribute default_namespace.
            Note this is not related to the default namespace of the document.
        owner: if True, call freeNode/Doc at destruction time
        """
        self._doc = None
        if isinstance(xml, StringTypes):
            xml = libxml2.parseMemory(xml, len(xml))
            owner = True
        if isinstance(xml, libxml2.xmlNode):
            if xml.type == 'element':
                self._elm = xml
            elif xml.type == 'document_xml':
                self._doc = xml
                self._elm = xml.getRootElement()
            else:
                assert False, 'xmlNode type %s not supported' % xml.type
        else:
            assert False, '%s not supported' % type(xml)
        self._owner = owner
        self.default_namespace = default_namespace

    elements = property(lambda self: self[:],
        doc='list of child elements')
    name = property(lambda self: (_ns(self._elm), self._elm.name),
        doc='element name')
    content = property(lambda self: self._elm.getContent(),
        doc='element content')

    def __del__(self):
        if self._owner:
            if self._doc:
                self._doc.freeDoc()
            else:
                try: self._elm.freeNode()
                except AttributeError: pass

    def _get_children(self):
        """-> list of this element's children (xmlElements)"""
        child_elements = []
        child = self._elm.children
        while child:
            if child.type == 'element':
                child_elements.append(child)
            child = child.next
        return child_elements

    def _get_children_by_name(self, name):
        """-> list of this element's children matching name (xmlElements)"""
        ns, name = _normalizeName(name, self.default_namespace)
        child_elements = []
        child = self._elm.children
        while child:
            if child.type == 'element' and child.name == name and \
                    _ns(child) == ns:
                child_elements.append(child)
            child = child.next
        return child_elements

    def __repr__(self):
        """Representation suitable for eval"""
        args = ''
        if self.default_namespace:
            args += ', default_namespace=%s' % repr(self.default_namespace)
        return '%s("""%s"""%s)' % (_qualified_class_name(self),
            self.serialize(standalone=True), args)

    def __str__(self):
        """Informal representation"""
        return self.serialize()

    def __call__(self, name):
        """Attribute access"""
        ns, t_name = _normalizeName(name)
        val = self._elm.nsProp(t_name, ns)
        if not val:
            raise KeyError(name)
        return val

    def __getitem__(self, key):
        """Child element access by index or name

        A name can be a string or a (namespace, name) tuple.

        Slices will return a list of elements.  Use [name:] to get all
        children with given name.  Negative indices are supported.
        """
        def new_child(x): return self.__class__(x, self.default_namespace)
        if isinstance(key, int):
            return new_child(self._get_children()[key])
        if isinstance(key, (StringTypes, tuple)):
            try:
                return new_child(self._get_children_by_name(key)[0])
            except IndexError:
                raise KeyError(key)
        if isinstance(key, slice):
            is_name = isinstance(key.start, (StringTypes, tuple)) and \
                key.stop is None and key.step is None
            if is_name:
                return map(new_child, self._get_children_by_name(key.start))
            else:
                return map(new_child, self._get_children()[key])
        raise TypeError('invalid key type')

    def has_element(self, name):
        """-> True if named child element exists"""
        try:
            self[name]
            return True
        except KeyError:
            return False

    def has_attribute(self, name):
        """-> True if named attribute exists"""
        try:
            self(name)
            return True
        except KeyError:
            return False

    def serialize(self, indent=False, standalone=False):
        """
        standalone: if True, include all namespace declarations
        """
        if not standalone:
            return self._elm.serialize(format=indent)
        else:
            copy = self._elm.copyNode(True)
            result = copy.serialize(format=indent)
            copy.freeNode()
            return result


#
# Derivations
#

class NameGenL(Namespace):
    """Name generation with translation of native Language constructs

    Also considered Lethal.
    """

    __slots__ = ['_underbar_to_dash']

    def __init__(self, uri, underbar_to_dash=False):
        Namespace.__init__(self, uri)
        self._underbar_to_dash = underbar_to_dash

    def __getattr__(self, name):
        if self._underbar_to_dash:
            return self[name.replace('_', '-')]
        else:
            return self[name]


class CachingElement(Element):
    """Caching version of Element

    Caching implementations of _get_children and _get_children_by_name,
    two often executed O(n) functions, are provided.
    """

    __slots__ = ['_cached_children', '_cached_children_by_name']

    def __init__(self, *args, **keyword_args):
        Element.__init__(self, *args, **keyword_args)
        # a list of children
        self._cached_children = None
        # a map of names to lists of children
        self._cached_children_by_name = None

    def _get_children(self):
        if not self._cached_children:
            self._cached_children = Element._get_children(self)
        return self._cached_children

    def _get_children_by_name(self, name):
        n_name = _normalizeName(name, self.default_namespace)
        if not self._cached_children_by_name:
            children_map = {}
            child = self._elm.children
            while child:
                if child.type == 'element':
                    child_name = (_ns(child), child.name)
                    children_map.setdefault(child_name, []).append(child)
                child = child.next
            self._cached_children_by_name = children_map
        return self._cached_children_by_name[n_name]


#
# testing
#

def _test():
    d = Element("""\
<doc version="2.7182818284590451"
        xmlns="http://example.org/bar"
        xmlns:dc="http://purl.org/dc/elements/1.1/"
        xmlns:bbc="http://example.org/bbc">
    <author><name>John Polk</name> and <name>John Palfrey</name></author>
    <dc:creator>John Polk</dc:creator>
    <dc:creator>John Palfrey</dc:creator>
    <bbc:show bbc:station="4">Buffy</bbc:show>
</doc>
""", default_namespace='http://example.org/bar')
    bar = NameGenL('http://example.org/bar')
    bbc = NameGenL('http://example.org/bbc')
    dc = NameGenL('http://purl.org/dc/elements/1.1/')
    print d
    print repr(d)
    print d('version')
    print d.name
    print d.elements
    print d[0]
    print d[-1]
    print d['author']
    print d['author'].serialize(standalone=True)
    assert repr(eval(repr(d['author']))) == repr(d['author'])
    print d[bar.author]
    print d[bar.author].content
    print d[dc.creator]
    print d[dc.creator:]
    print d[1:]
    print d[bbc.show](bbc.station)

    d2 = Element('<foo><bar/></foo>')
    print d2.serialize(indent=True)
    print d2['bar']

    d3 = Element('<foo><bar-bar/></foo>')
    ns3 = NameGenL(None, underbar_to_dash=True)
    print d3[ns3.bar_bar]

    d4 = CachingElement(
        '<foo><bar><baz>hello</baz><baz>goodbye</baz></bar></foo>')
    print d4.name
    print d4['bar'].content
    print d4['bar'].elements
    print d4['bar'][0]
    print d4['bar'][-1]
    print d4.serialize(indent=True)

def _main():
    libxml2.debugMemory(True)
    _test()
    libxml2.cleanupParser()
    leaked = libxml2.debugMemory(True)
    if leaked:
        print "*** libxml2 memory leak %d bytes ***" % leaked
        libxml2.dumpMemory()

if __name__ == '__main__':
    _main()
