# Copyright (C) 2004  Anthony Carrico

"""Basic xmldsig + OpenPGP support"""
from giftfile import openpgp
from giftfile import funkxml
from giftfile.funkxml import ExpName, Element, Attribute, \
     Root, NsDecl, Text, Node
from giftfile._data_util import to_base64, from_base64, \
     sha1, sha1_base64, to_funkxml_root_element, \
     hex_to_base64, base64_to_hex, check_string, \
     check_uri, fetch_sha1_base64, trap_key_error, XmlObject
from giftfile._context import Context, SharedResource
from giftfile import relaxng

import libxml2
import sys

def warning(message):
    print >> sys.stderr, 'WARNING: %s' % message

class Error(Exception):
    def __init__(self, message=''):
        Exception.__init__(self, message)

class SigningError(Error):
    def __init__(self):
        Error.__init__(self)

class XmlValidationError(Error):
    def __init__(self):
        Error.__init__(self)

class SignatureDigestMismatch(Error):
    def __init__(self):
        Error.__init__(self)

# Standard identifiers:
_ns_ds = 'http://www.w3.org/2000/09/xmldsig#'
_ns_openpgp = 'http://giftfile.org/2004/02/openpgp#'
_algo_id_sha1 = "http://www.w3.org/2000/09/xmldsig#sha1"
_algo_id_c14n = "http://www.w3.org/2001/10/xml-exc-c14n#"
_algo_id_dsa_sha1 = "http://www.w3.org/2000/09/xmldsig#dsa-sha1"
_algo_id_enved_sig = "http://www.w3.org/2000/09/xmldsig#enveloped-signature"

# Custom identifiers:
_id_pgp_fingerprint = "http://giftfile.org/2004/02/openpgp#fingerprint"
_id_pgp_private_key_block = (
    "http://giftfile.org/2004/02/openpgp#private-key-block")
_id_pgp_pub_key_block = (
    "http://giftfile.org/2004/02/openpgp#public-key-block")
_algo_id_pgp_message = "http://giftfile.org/2004/02/openpgp#message"
_algo_id_pgp_sig = "http://giftfile.org/2004/02/openpgp#signature"

# Namespace declarations:
_ds = NsDecl("ds", _ns_ds)
_openpgp = NsDecl("ds", _ns_openpgp)

def get_ns_decl_ds():
    return _ds

def get_ns_decl_openpgp():
    return _openpgp

# Elements:
_elem_method_sha1 = Element(
    None, _ds('DigestMethod'),
    (), (Attribute(None, 'Algorithm', _algo_id_sha1),))

def get_digest_method_sha1():
    return _elem_method_sha1

_elem_method_c14n = Element(
    None, _ds('CanonicalizationMethod'),
    (), (Attribute(None, 'Algorithm', _algo_id_c14n),))

_elem_method_pgp_sig = Element(
    None, _ds('SignatureMethod'),
    (), (Attribute(None, 'Algorithm', _algo_id_pgp_sig),))

_elem_trans_c14n = Element(
    None, _ds('Transform'),
    (), (Attribute(None, 'Algorithm', _algo_id_c14n),))

_elem_trans_enved_sig = Element(
    None, _ds('Transform'),
    (), (Attribute(None, 'Algorithm', _algo_id_enved_sig),))

# Accessors

def sig__signed_info(sig):
    return sig.content_map[_ds('SignedInfo')][0]

def sig__signature_value(sig):
    return from_base64(sig.content_map[_ds('SignatureValue')][0].string_value)

def signed_info__refs(signed_info):
    return signed_info.content_map[_ds('Reference')]

def ref__digest_value(ref):
    return from_base64(ref.content_map[_ds('DigestValue')][0].string_value)

def get_message_digest(sig):
    """-> binary digest value

    This function allows an application to access the message digest
    of a signature without knowing the details of xmldsig.  It assumes
    a valid signature with exactly one reference.

    ISSUE: roll this into validate functions?
        Validate functions should probably return a structure encapsulating
        more info about the signature.  For example:

            sig_info = validate_enveloped_openpgp_custom(...)
            print sig_info.signers
            print sig_info.message_digest_algo, sig_info.message_digest
    """

    if not isinstance(sig, Node):
        # Support libxml2:
        sig = funkxml.from_libxml2(sig, is_orphan=True)

    refs = signed_info__refs(sig__signed_info(sig))
    # This procedure assumes the signature has exactly one reference:
    assert (len(refs) == 1)
    return ref__digest_value(refs[0])

class KeyInfo(object):
    """OpenPGP fingerprint as xmldsig KeyInfo"""

    base64 = property(
        lambda self: self._rep,
        doc='OpenPGP fingerprint as base64 encoded string.')

    hex = property(
        lambda self: base64_to_hex(self._rep),
        doc='OpenPGP fingerprint as hex encoded string.')

    binary = property(
        lambda self: from_base64(self._rep),
        doc='OpenPGP fingerprint as binary string.')

    funkxml = property(
        lambda self: Element.auto(
        None, _ds('KeyInfo'),
        Element.auto(None, _openpgp('Fingerprint'),
                     self.base64)),
        doc='KeyInfo as funkxml Element.')

    def make_from_hex(fingerprint):
        return KeyInfo(hex_to_base64(fingerprint), True)
    make_from_hex = staticmethod(make_from_hex)

    def make_from_base64(fingerprint):
        return KeyInfo(fingerprint, True)
    make_from_base64 = staticmethod(make_from_base64)

    def make_from_binary(fingerprint):
        return KeyInfo(to_base64(fingerprint), True)
    make_from_binary = staticmethod(make_from_binary)

    def make_from_funkxml(element):
        return KeyInfo.make_from_base64(
            element.content_map[_openpgp('Fingerprint')][0].string_value)
    make_from_funkxml = staticmethod(make_from_funkxml)

    def __init__(self, rep, really=False):
        '''Please use one of the make* constructors.'''
        assert really
        self._rep = check_string(rep)

class Reference(object):
    def __get_funkxml(self):
        if self.transforms == None:
            transforms = None
        else:
            transforms = Element.auto(
                None, _ds('Transforms'),
                map(lambda transform:
                    Element.auto(None, _ds('Transform'),
                                 Attribute(None, 'Algorithm', transform)),
                    self.transforms))
        if self.uri == None:
            uri = None
        else:
            uri = Attribute(None, 'URI', self.uri),

        if self.id == None:
            id = None
        else:
            id = Attribute(None, 'Id', self.uri),

        if self.type == None:
            type = None
        else:
            type = Attribute(None, 'Type', self.uri),

        return Element.auto(
            None, _ds('Reference'),
            transforms,
            Element.auto(None, _ds('DigestMethod'),
                         Attribute(None, 'Algorithm', self.digest_method)),
            Element.auto(None, _ds('DigestValue'), self.digest_value),
            uri, id, type)

    funkxml = property(__get_funkxml)

    def fetch_and_make(fetch_uri,
                       transforms=None, digest_method=_algo_id_sha1,
                       uri=None, id=None, type=None):
        assert digest_method == _algo_id_sha1
        assert transforms == None
        return Reference(transforms, digest_method,
                         fetch_sha1_base64(fetch_uri),
                         uri, id, type, True)
    fetch_and_make = staticmethod(fetch_and_make)

    def make_from_value(
        value,
        transforms=None, digest_method=_algo_id_sha1,
        uri=None, id=None, type=None):
        assert digest_method == _algo_id_sha1
        assert transforms == None
        digest_value = sha1_base64(check_string(value))
        return Reference(transforms, digest_method, digest_value,
                         uri, id, type, True)
    make_from_value = staticmethod(make_from_value)

    def make_from_digest_hex(
        digest_value,
        transforms=None, digest_method=_algo_id_sha1,
        uri=None, id=None, type=None):
        return Reference(transforms, digest_method,
                         hex_to_base64(digest_value),
                         uri, id, type, True)
    make_from_digest_hex = staticmethod(make_from_digest_hex)

    def make_from_digest_base64(
        digest_value,
        transforms=None, digest_method=_algo_id_sha1,
        uri=None, id=None, type=None):
        return Reference(transforms, digest_method,
                         digest_value,
                         uri, id, type, True)
    make_from_digest_base64 = staticmethod(make_from_digest_base64)

    def make_from_digest_binary(
        digest_value,
        transforms=None, digest_method=_algo_id_sha1,
        uri=None, id=None, type=None):
        return Reference(transforms, digest_method,
                         to_base64(digest_value),
                         uri, id, type, True)
    make_from_digest_binary = staticmethod(make_from_digest_binary)

    def make_from_funkxml(element):
        return Reference(
            trap_key_error(lambda: map(
            lambda self: self.attrs['Algorithm'].string_value,
            element.content_map[_ds('Transforms')][0]
            .content_map[_ds('Transform')])),
            element.content_map[_ds('DigestMethod')][0]
            .attrs['Algorithm'].string_value,
            element.content_map[_ds('DigestValue')][0].string_value,
            trap_key_error(lambda: element.attrs['URI'].string_value),
            trap_key_error(lambda: element.attrs['Id'].string_value),
            trap_key_error(lambda: element.attrs['Type'].string_value),
            True)
    make_from_funkxml = staticmethod(make_from_funkxml)

    def __init__(self,
                 transforms, digest_method, digest_value,
                 uri, id, type,
                 really=False):
        '''Please use one of the make* constructors.'''
        assert(really)
        if transforms != None:
            transforms = tuple(transforms)
            if transforms == ():
                transforms = None
        assert ((transforms == None)
                or (transforms == (_algo_id_enved_sig,
                                   _algo_id_c14n)))
        self.transforms = transforms
        assert digest_method == _algo_id_sha1
        self.digest_method = digest_method
        self.digest_value = check_string(digest_value)
        if uri != None: check_uri(uri)
        self.uri = uri
        if id != None: check_string(id)
        self.id = id
        if type != None: check_string(type)
        self.type = type

class Signature(XmlObject):
    references = property(
        lambda self:
        map(Reference.make_from_funkxml,
            self.element.content_map[_ds('SignedInfo')][0]
            .content_map[_ds('Reference')]))

    _schema_ctxt = SharedResource(lambda: relaxng.fetch_ctxt(
        'http://giftfile.org/2004/02/xmldsig-custom.rng'))

    def __init__(self, xml):
        XmlObject.__init__(self, xml)

class enveloped_openpgp_custom(object):
    """Sign and verify xml documents as described in xmldsig-custom.rng."""

    _schema_ctxt = SharedResource(lambda: relaxng.fetch_ctxt(
        'http://giftfile.org/2004/02/xmldsig-custom.rng'))

    def __init__(self, pgp_ctxt):
        self.schema_ctxt = None
        self.pgp_ctxt = pgp_ctxt

    def sign (self, bogus_sig,
              signers = None,
              ref_id = None,
              ref_type = None,
              sig_id = None):

        """Return an enveloped signature of the bogus_sig's document.

        The bogus_sig can be a funkxml or libxml2 node. The original
        document is not changed.

        The reference's URI attribute is set to the empty string. This
        is appropriate when making an enveloped signature of an entire
        document. The actual node could be a Root or Element node.

        The document will be signed by the given signer(s) which can
        be KeyInfo instances or values recognized by the context. If
        signers is None, the context's defaults will be used.

        The reference is given ID and/or Type attributes if a ref_id
        and/or ref_type are given.

        The signature is given an ID attribute if a sig_id is given.

        The resulting signature is an orphan element, it should be
        inserted someplace in the given node."""

        libxml2_bogus_sig = None
        if isinstance(bogus_sig, libxml2.xmlNode):
            libxml2_bogus_sig = bogus_sig
            bogus_sig = funkxml.from_libxml2(bogus_sig)

        if signers != None:
            tmp_signers = []
            for s in signers:
                if isinstance(s, KeyInfo): s = s.hex
                tmp_signers.append(s)
            signers = tmp_signers

        # 3.1.1 Reference Generation

        # For each data object being signed:

        # 1. Apply the Transforms, as determined by the application,
        # to the data object.

        content = bogus_sig.parent.content.orphan
        index = bogus_sig.index
        trans_doc = bogus_sig.parent.with_content(
            content[0:index] + content[index+1:]).top

        # 2. Calculate the digest value over the resulting data
        # object.

        digest_value = Element(
            None, _ds('DigestValue'),
            (Text(sha1_base64(trans_doc.c14n_serialize(exclusive=True))),))

        # 3. Create a Reference element, including the (optional)
        # identification of the data object, any (optional) transform
        # elements, the digest algorithm and the DigestValue. (Note,
        # it is the canonical form of these references that are signed
        # in 3.1.2 and validated in 3.2.1 .)

        ref_uri = ''
        reference_attrs = [Attribute(None, 'URI', ref_uri)]

        if ref_id != None:
            reference_attrs.append(Attribute(None, 'Id', ref_id))

        if ref_type != None:
            reference_attrs.append(Attribute(None, 'Type', ref_type))

        reference = Element(None, _ds('Reference'),
                            (Element(None, _ds('Transforms'),
                                     (_elem_trans_enved_sig,
                                      _elem_trans_c14n,)),
                             _elem_method_sha1,
                             digest_value),
                            reference_attrs)

        # 3.1.2 Signature Generation

        # 1. Create SignedInfo element with SignatureMethod,
        # CanonicalizationMethod and Reference(s).

        signed_info = Element(None, _ds('SignedInfo'),
                              (_elem_method_c14n,
                               _elem_method_pgp_sig,
                               reference))

        # 2. Canonicalize and then calculate the SignatureValue over
        # SignedInfo based on algorithms specified in SignedInfo.

        sns, sig = self.pgp_ctxt.detach_sign(
            signed_info.orphan_with_auto_ns_decls.c14n_serialize(exclusive=True),
            signers)
        if not sig: raise SigningError

        # WARNING: not currently checking the signer's ids against the sns!
        if signers:
            if len(sns) != len(signers):
                raise Error('[when do I occur?]')
        else:
            if len(sns) != 1:
                raise Error('[when do I occur?]')

        signature_value = Element(None, _ds('SignatureValue'), (Text(to_base64(sig)),))

        # 3. Construct the Signature element that includes SignedInfo,
        # Object(s) (if desired, encoding may be different than that
        # used for signing), KeyInfo (if required), and
        # SignatureValue.

        signature_attrs = []

        if sig_id != None:
            signature_attrs.append(Attribute(None, 'Id', sig_id))

        orphan_sig = Element(None, _ds('Signature'),
                             (signed_info, signature_value),
                             signature_attrs).orphan_with_auto_ns_decls

        if libxml2_bogus_sig == None:
            return orphan_sig.replace(bogus_sig)
        else:
            sig = orphan_sig.libxml2_replace(libxml2_bogus_sig)
            libxml2_bogus_sig.freeNode()
            return sig

    def validate(self, sig, validate_xml=True):
        """Validate an enveloped signature.

        The sig can be a funkxml or libxml2 node.

        Raises: XmlValidationError, SignatureDigestMismatch

        The result is like openpgp.Context.detach_verify: a list of a list
        of opengpg.SigGood, opengpg.SigBad, and/or opengpg.SigError
        elements, one for each signer.
        """

        if isinstance(sig, libxml2.xmlNode):
            sig = funkxml.from_libxml2(sig)

        if validate_xml:
            if self.schema_ctxt == None:
                self.schema_ctxt = self._schema_ctxt()
            if not sig.relaxng_validate(self.schema_ctxt):
                raise XmlValidationError

        # 3.2.1 Reference Validation

        # 1. Canonicalize the SignedInfo element based on the
        # CanonicalizationMethod in SignedInfo.

        orig_signed_info = sig__signed_info(sig)
        c14n_signed_info = (orig_signed_info
                            .orphan_with_auto_ns_decls
                            .c14n_serialize(exclusive=True))
        signed_info = funkxml.parse_memory(c14n_signed_info).element

        # 2. For each Reference in SignedInfo:

        ref = signed_info__refs(signed_info)[0]

        #  1. Obtain the data object to be digested. (For example, the
        #  signature application may dereference the URI and execute
        #  Transforms provided by the signer in the Reference element, or
        #  it may obtain the content through other means such as a local
        #  cache.)

        # Remove the sig from the document:
        content = sig.parent.content.orphan
        trans_doc = sig.parent.with_content(
            content[0:sig.index] + content[sig.index+1:]).top

        #  2. Digest the resulting data object using the DigestMethod
        #  specified in its Reference specification.

        digest_value = sha1(trans_doc.c14n_serialize(exclusive=True))

        #  3. Compare the generated digest value against DigestValue in
        #  the SignedInfo Reference; if there is any mismatch, validation
        #  fails.

        if digest_value != ref__digest_value(ref):
            raise SignatureDigestMismatch

        # 3.2.2 Signature Validation

        # 1. Obtain the keying information from KeyInfo or from an
        # external source.

        # 2. Obtain the canonical form of the SignatureMethod using the
        # CanonicalizationMethod and use the result (and previously
        # obtained KeyInfo) to confirm the SignatureValue over the
        # SignedInfo element.

        # Additionally, the SignatureMethod URI may have been altered by
        # the canonicalization of SignedInfo (e.g., absolutization of
        # relative URIs) and it is the canonical form that MUST be
        # used. However, the required canonicalization [XML-C14N] of this
        # specification does not change URIs.

        signature_value = sig__signature_value(sig)

        openpgp_sigs = self.pgp_ctxt.detach_verify(
            signature_value, c14n_signed_info)

        return openpgp_sigs

def funkxml_example(sig_ctxt, source):
    doc = funkxml.parse_memory(source)
    bogus_sig = doc.element.content_map['Signature'][0]

    print "Before:"
    print doc

    sig = sig_ctxt.sign(bogus_sig,
                        # The rest are all optional:
                        # signers = ["jane-doe", "john-doe"],
                        sig_id = "hello-sig",
                        ref_id = "hello-ref",
                        ref_type = "http://bogus/example")

    print
    print "Sig:"
    print "NOTE: the indentation is for good looks, but it breaks the signature!"
    print sig.serialize(indent=True)

    signed_doc = sig.top

    print
    print "After (no indentation):"
    print signed_doc.serialize(indent=False)

    print
    print "Validate:"
    print

    for v in sig_ctxt.validate(sig):
        print
        print "Verify: " + repr(v.__dict__)

def libxml2_example(sig_ctxt, source):
    doc = None
    try:
        doc = libxml2.parseMemory(source, len(source))
        bogus_sig = doc.children.children.next

        print "Before:"
        print doc.serialize()

        sig = sig_ctxt.sign(bogus_sig,
                            # The rest are all optional:
                            # signers = ["jane-doe", "john-doe"],
                            sig_id = "hello-sig",
                            ref_id = "hello-ref",
                            ref_type = "http://bogus/example")

        print
        print "Sig:"
        print "NOTE: the indentation is for good looks, but it breaks the signature!"
        print sig.serialize(format=True)

        print
        print "After (no indentation):"
        print doc.serialize(format=False)

        print
        print "Validate:"
        print

        for v in sig_ctxt.validate(sig):
            print
            print "Verify: " + repr(v.__dict__)

    finally:
        if doc != None: doc.freeDoc()

def examples():
    import os

    funkxml.init_libxml2_debug()

    from os.path import expanduser, exists

    # NOTE: Use a bogus keyring if available:
    testing_gpg_homedir = expanduser('~/tmp/gpg')
    if exists(testing_gpg_homedir):
        gpg_homedir = testing_gpg_homedir
    else:
        gpg_homedir = None

    pgp_ctxt = openpgp.Context(homedir=gpg_homedir)
    sig_ctxt = enveloped_openpgp_custom(pgp_ctxt)

    source = """<test><greet>
hello
</greet><Signature/><greet>
world
</greet></test>"""

    funkxml_example(sig_ctxt, source)

    libxml2_example(sig_ctxt, source)

    sig_ctxt = None

    funkxml.quit_libxml2_debug()

if __name__ == "__main__":
    examples()
